# This file is part of the gmx-top4py project.
#
# The gmx-top4py project is based on or includes code from:
#    kimmdy (https://github.com/graeter-group/kimmdy/tree/main)
#    Copyright (C) graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# Modifications and additional code:
#    Copyright (C) 2025 graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import textwrap
from pathlib import Path
from typing import Optional

from gmx_top4py.constants import FFFUNC
from gmx_top4py.parsing import read_top
from gmx_top4py.topology.atomic import (
    AngleId,
    AngleType,
    AtomId,
    AtomType,
    BondId,
    BondType,
    DihedralType,
    ImproperDihedralId,
    NonbondParamType,
    ProperDihedralId,
    ResidueType,
)
from gmx_top4py.topology.utils import get_top_section
from gmx_top4py.utils import get_gmx_dir

logger = logging.getLogger(__name__)


class FF:
    """Container for parsed forcefield data.

    Also see <https://manual.gromacs.org/current/reference-manual/topologies/topology-file-formats.html#topology-file>
    """

    def __init__(
        self,
        top: dict,
        residuetypes_path: Optional[Path] = None,
        gromacs_alias: str = "gmx",
    ):
        self.atomtypes: dict[AtomId, AtomType] = {}
        self.bondtypes: dict[BondId, BondType] = {}
        self.angletypes: dict[AngleId, AngleType] = {}
        self.proper_dihedraltypes: dict[ProperDihedralId, DihedralType] = {}
        self.improper_dihedraltypes: dict[ImproperDihedralId, DihedralType] = {}
        self.residuetypes: dict[str, ResidueType] = {}
        self.nonbond_params: dict[BondId, NonbondParamType] = {}

        ffdir: Optional[Path] = top["ffdir"]

        self.gmxdir = get_gmx_dir(gromacs_alias)
        if self.gmxdir is None:
            logger.warning(f"Could not find gromacs data directory for {gromacs_alias}")

        if defaults := get_top_section(top, "defaults"):
            self.defaults = defaults

        atomtypes = get_top_section(top, "atomtypes")
        if atomtypes is not None:
            for l in atomtypes:
                atomtype = AtomType.from_top_line(l)
                self.atomtypes[atomtype.type] = atomtype
        bondtypes = get_top_section(top, "bondtypes")
        if bondtypes is not None:
            for l in bondtypes:
                bondtype = BondType.from_top_line(l)
                self.bondtypes[(bondtype.i, bondtype.j)] = bondtype

        nonbond_params = get_top_section(top, "nonbond_params")
        if nonbond_params is not None:
            for l in nonbond_params:
                nonbond_param = NonbondParamType.from_top_line(l)
                self.nonbond_params[(nonbond_param.i, nonbond_param.j)] = nonbond_param

        angletypes = get_top_section(top, "angletypes")
        if angletypes is not None:
            for l in angletypes:
                angletype = AngleType.from_top_line(l)
                self.angletypes[(angletype.i, angletype.j, angletype.k)] = angletype

        dihedraltypes = get_top_section(top, "dihedraltypes")
        if dihedraltypes is not None:
            for l in dihedraltypes:
                dihedraltype = DihedralType.from_top_line(l)
                # proper dihedrals can be defined multiple times
                # with a different phase
                if dihedraltype.funct == FFFUNC["mult_improper_dihedral"]:
                    self.improper_dihedraltypes[
                        (dihedraltype.i, dihedraltype.j, dihedraltype.k, dihedraltype.l)
                    ] = dihedraltype
                else:
                    # e.g. proper dihedrals with dihedraltype.funct == "9":
                    if (
                        self.proper_dihedraltypes.get(
                            (
                                dihedraltype.i,
                                dihedraltype.j,
                                dihedraltype.k,
                                dihedraltype.l,
                                dihedraltype.periodicity,
                            )
                        )
                        is None
                    ):
                        self.proper_dihedraltypes[
                            (
                                dihedraltype.i,
                                dihedraltype.j,
                                dihedraltype.k,
                                dihedraltype.l,
                                dihedraltype.periodicity,
                            )
                        ] = dihedraltype

        if residuetypes_path:
            logger.debug(f"Using specified residuetypes file: {residuetypes_path}")
            residuetypes_paths = [residuetypes_path]
        else:
            logger.debug("Trying to use default amber protein residuetypes file.")
            if ffdir is None:
                logger.warning("ffdir is None. No residuetypes will be parsed.")
                return
            residuetypes_paths = [
                ffdir / x for x in ["rna.rtp", "dna.rtp", "aminoacids.rtp"]
            ]

        for residuetypes_path in residuetypes_paths:
            if not residuetypes_path.exists():
                logger.warning(f"{residuetypes_path} not found in ffdir.")
                continue
            residuetypes_dict = read_top(residuetypes_path, use_gmx_dir=False)
            for k, v in residuetypes_dict.items():
                if k.startswith("BLOCK") or k in ["bondedtypes", "ffdir", "define"]:
                    continue
                if not v.get("subsections"):
                    raise AssertionError(f"key {k} has no subsections, only {v}.")
                self.residuetypes[k] = ResidueType.from_section(k, v["subsections"])

    def __str__(self) -> str:
        return textwrap.dedent(
            f"""
        ForceField parameters with
        {len(self.atomtypes)} atomtypes,
        {len(self.bondtypes)} bondtypes,
        {len(self.angletypes)} angletypes,
        {len(self.proper_dihedraltypes)} proper dihedraltypes
        {len(self.improper_dihedraltypes)} improper dihedraltypes
        {len(self.residuetypes)} residuetypes
        """
        )

    def __repr__(self) -> str:
        return f"FF({self.__dict__})"

    def _repr_pretty_(self, p, _):
        """A __repr__ for ipython.

        This whill be used if just the name of the object is entered in the ipython shell
        or a jupyter notebook.

        p is an instance of [IPython.lib.pretty.RepresentationPrinter](https://ipython.org/ipython-doc/3/api/generated/IPython.lib.pretty.html#IPython.lib.pretty.PrettyPrinter)
        """
        p.text(str(self))



# This file is part of the gmx-top4py project.
#
# The gmx-top4py project is based on or includes code from:
#    kimmdy (https://github.com/graeter-group/kimmdy/tree/main)
#    Copyright (C) graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# Modifications and additional code:
#    Copyright (C) 2025 graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""Base classes and basic instances for parameterizing topologies."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from gmx_top4py.topology.topology import Topology


class Parameterizer(ABC):
    def __init__(self, **kwargs):
        self.type_scheme = dict()

    @abstractmethod
    def parameterize_topology(
        self, current_topology: Topology, focus_nrs: Optional[set[str]]
    ) -> Topology:
        pass


class BasicParameterizer(Parameterizer):
    """reconstruct base force field state"""

    def parameterize_topology(
        self, current_topology: Topology, focus_nrs: Optional[set[str]] = None
    ) -> Topology:
        """Do nothing,
        all necessary actions should already have happened in bind_bond and break_bond of Topology
        """
        return current_topology



# This file is part of the gmx-top4py project.
#
# The gmx-top4py project is based on or includes code from:
#    kimmdy (https://github.com/graeter-group/kimmdy/tree/main)
#    Copyright (C) graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# Modifications and additional code:
#    Copyright (C) 2025 graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations  # for 3.7 <= Python version < 3.10

import logging
import re
from itertools import permutations
from typing import TYPE_CHECKING, Any, Callable, Optional

from gmx_top4py.constants import ION_NAMES, SOLVENT_NAMES
from gmx_top4py.topology.atomic import AtomId, MoleculeTypeHeader

if TYPE_CHECKING:
    # from config import Config
    from gmx_top4py.topology.atomic import Atom, AtomicType, AtomicTypes
    from gmx_top4py.topology.topology import Topology

logger = logging.getLogger(__name__)


def get_top_section(
    top: dict, name: str, moleculetype: Optional[str] = None
) -> Optional[list[list]]:
    """Get content of a section from a topology dict.

    Resolves any `#ifdef` statements by check in the top['define'] dict
    and chooses the 'content' or 'else_content' depending on the result.
    """
    if moleculetype is not None:
        parent_name = f"moleculetype_{moleculetype}"
        parent_section = top.get(parent_name)
        if parent_section is None:
            logger.warning(f"topology does not contain moleculetype {moleculetype}")
            return None
        section = parent_section["subsections"].get(name)
    else:
        section = top.get(name)

    if section is None:
        logger.debug(
            f"Topology does not have {name}. "
            "If you expet to find this section, check that "
            "the forcefield is in the current working directory config.cwd."
        )
        return None
    condition = section.get("condition")
    if condition is not None:
        condition_type = condition.get("type")
        condition_value = condition.get("value")
        if condition_type == "ifdef":
            if condition_value in top["define"].keys():
                return section.get("content")
            else:
                return section.get("else_content")
        elif condition_type == "ifndef":
            if condition_value not in top["define"].keys():
                return section.get("content")
            else:
                return section.get("else_content")
        else:
            raise NotImplementedError(
                f"condition type {condition_type} is not supported"
            )
    return section.get("content")


def get_moleculetype_header(
    top: dict, moleculetype: str
) -> Optional[MoleculeTypeHeader]:
    """Get content of the header of a moleculetype from a topology dict.

    Resolves any `#ifdef` statements by check in the top['define'] dict
    and chooses the 'content' or 'else_content' depending on the result.
    """
    section = top.get(moleculetype)
    if section is None:
        logger.warning(f"topology does not contain moleculetype {moleculetype}")
        return None

    condition = section.get("condition")
    if condition is not None:
        condition_type = condition.get("type")
        condition_value = condition.get("value")
        if condition_type == "ifdef":
            if condition_value in top["define"].keys():
                return section.get("content")
            else:
                return section.get("else_content")
        elif condition_type == "ifndef":
            if condition_value not in top["define"].keys():
                return section.get("content")
            else:
                return section.get("else_content")
        else:
            raise NotImplementedError(
                f"condition type {condition_type} is not supported"
            )
    name, nrexcl = section.get("content")[0]
    if name is None:
        logger.info(
            f"name not found in moleculetype {moleculetype}. Defaulting to Unknown."
        )
        name = "Unknown"
    if nrexcl is None:
        logger.info(
            f"nrexcl not found in moleculetype {moleculetype}. Defaulting to 3."
        )
        nrexcl = "3"
    return MoleculeTypeHeader(name=name, nrexcl=nrexcl)


def get_moleculetype_atomics(top: dict, moleculetype: str) -> Optional[dict]:
    """Get content of subsections (atoms/bonds/angles etc.) of a moleculetype from a topology dict.

    Resolves any `#ifdef` statements by check in the top['define'] dict
    and chooses the 'content' or 'else_content' depending on the result.
    """
    section = top.get(moleculetype)
    if section is None:
        logger.warning(f"topology does not contain moleculetype {moleculetype}")
        return None

    subsections = section["subsections"]
    atomics = {}
    for k, v in subsections.items():
        condition = v.get("condition")
        if condition is not None:
            condition_type = condition.get("type")
            condition_value = condition.get("value")
            if condition_type == "ifdef":
                if condition_value in top["define"].keys():
                    atomics[k] = v.get("content")
                else:
                    atomics[k] = v.get("else_content")
            elif condition_type == "ifndef":
                if condition_value not in top["define"].keys():
                    atomics[k] = v.get("content")
                else:
                    atomics[k] = v.get("else_content")
            else:
                raise NotImplementedError(
                    f"condition type {condition_type} is not supported"
                )
        else:
            atomics[k] = v["content"]

    return atomics


def get_protein_section(top: dict, name: str) -> Optional[list[list]]:
    """Get content of a section in the first moleculetype (protein) from a topology dict."""
    return get_top_section(top, name, moleculetype="Protein")


def get_selected_section(
    top: dict, name: str, selected_moleculetype: str
) -> Optional[list[list]]:
    """Get content of a section in the selected moleculetype from a topology dict."""
    return get_top_section(top, name, moleculetype=selected_moleculetype)


def set_top_section(
    top: dict, name: str, value: list, moleculetype: Optional[str] = None
) -> Optional[list[list]]:
    """Set content of a section from a topology dict.

    Resolves any `#ifdef` statements by check in the top['define'] dict
    and chooses the 'content' or 'else_content' depending on the result.
    """
    if moleculetype is not None:
        parent_name = f"moleculetype_{moleculetype}"
        parent_section = top.get(parent_name)
        if parent_section is None:
            raise ValueError(f"topology does not contain moleculetype {moleculetype}")
        section = parent_section["subsections"].get(name)
    else:
        section = top.get(name)

    if section is None:
        m = f"topology does not contain section {name}"
        logger.warning(m)
        return None
    condition = section.get("condition")
    if condition is not None:
        condition_type = condition.get("type")
        condition_value = condition.get("value")
        if condition_type == "ifdef":
            if condition_value in top["define"].keys():
                section["content"] = value
            else:
                section["else_content"] = value
        elif condition_type == "ifndef":
            if condition_value not in top["define"].keys():
                section["content"] = value
            else:
                section["else_content"] = value
        else:
            raise NotImplementedError(
                f"condition type {condition_type} is not supported"
            )
    section["content"] = value


def attributes_to_list(obj) -> list[str]:
    attrs = []
    for k, v in obj.__dict__.items():
        if k in ["bound_to_nrs", "is_radical", "id", "id_sym"]:
            continue
        if v in [None, ""]:
            continue
        if isinstance(v, (list, tuple)):
            attrs.extend(v)
        if isinstance(v, str):
            attrs.append(v)
    return attrs


def is_not_none(x) -> bool:
    return x is None


def get_by_permutations(d: dict, key) -> Optional[Any]:
    for k in permutations(key):
        value = d.get(k, None)
        if value is not None:
            return value
    return None


def match_atomic_item_to_atomic_type(
    id: list[str], types: AtomicTypes, periodicity: str = ""
) -> Optional[AtomicType]:
    def escape_re_atomtypes(s: str) -> str:
        """
        escape special regex characters
        that can appear in a forcefield
        """
        return s.replace("*", "STAR").replace("+", "PLUS")

    id = [escape_re_atomtypes(s) for s in id]
    id_sym = id[::-1]
    id_str = "---".join(id)
    id_sym_str = "---".join(id_sym)
    if periodicity:
        id[-1] += ":::" + periodicity
        id_sym[-1] += ":::" + periodicity
        id_str += ":::" + periodicity
        id_sym_str += ":::" + periodicity
    result = None
    longest_match = 0
    for _, atomic_type in types.items():
        for key in [atomic_type.id, atomic_type.id_sym]:
            key = escape_re_atomtypes(key).replace("X", ".*")
            keys = key.split("---")
            # early return exact match
            if key == id_str or key == id_sym_str:
                return atomic_type
            matches = [re.match(pattern, s) for pattern, s in zip(keys, id)]
            if all(matches):
                # favor longer (=more specific) and later matches
                if len(key) >= longest_match:
                    longest_match = len(key)
                    result = atomic_type

    return result


def increment_field(l: list[str], i: int, n: int):
    l[i] = str(int(l[i]) + n)
    return l


def is_not_solvent_or_ion(name: str) -> bool:
    """Returns whether a moleculetype name is not solvent or ion."""
    return name.lower() not in [x.lower() for x in SOLVENT_NAMES + ION_NAMES]


def get_is_selected_moleculetype_f(
    selected: list[str], deselected: list[str]
) -> Callable[[str], bool]:
    """Returns whether a moleculetype name is selected or not.

    Per default, solvents and inorganic ions are deselected.
    """
    default_deselected = [x.lower() for x in SOLVENT_NAMES + ION_NAMES]
    deselected = [s.lower() for s in deselected]
    selected = [s.lower() for s in selected]

    def f(name: str) -> bool:
        lower_name = name.lower()
        return lower_name not in deselected and (
            lower_name not in default_deselected or lower_name in selected
        )

    return f


def get_residue_by_bonding(atom: Atom, atoms: dict[AtomId, Atom]) -> dict[AtomId, Atom]:
    """Get the residue of an atom by its bonding.

    Avoids traversing the whole topology.

    Parameters
    ----------
    atom
        Atom of the residue
    atoms
        All atoms of a topology

    Returns
    -------
        Atoms of the residue
    """

    def rec(
        atom: Atom, atoms: dict[str, Atom], residue: dict[str, Atom]
    ) -> dict[str, Atom]:
        if atom.nr in residue.keys():
            # already visited
            return residue
        residue[atom.nr] = atom
        for nr in atom.bound_to_nrs:
            if atoms[nr].resnr == atom.resnr:
                rec(atoms[nr], atoms, residue)
        return residue

    residue = {}
    rec(atom, atoms, residue)
    return residue



# This file is part of the gmx-top4py project.
#
# The gmx-top4py project is based on or includes code from:
#    kimmdy (https://github.com/graeter-group/kimmdy/tree/main)
#    Copyright (C) graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# Modifications and additional code:
#    Copyright (C) 2025 graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Atomic datatypes for the topology such as Atom, Bond, Angle, Dihedral, etc.
The order of the fields comes from the gromacs topology file format.
See [gromacs manual](https://manual.gromacs.org/current/reference-manual/topologies/topology-file-formats.html#topology-file)
"""

import logging
from dataclasses import dataclass, field
from typing import Optional, Union

from gmx_top4py.constants import FFFUNC
from gmx_top4py.utils import field_or_none

logger = logging.getLogger(__name__)


@dataclass
class MoleculeTypeHeader:
    name: str
    nrexcl: str


@dataclass()
class Atom:
    """Information about one atom

    A class containing atom information as in the atoms section of the topology.
    An atom keeps a list of which atoms it is bound to and its radical state.

    From gromacs topology:
    ; nr type resnr residue atom cgnr charge mass typeB chargeB massB
    """

    nr: str
    type: str
    resnr: str
    residue: str
    atom: str
    cgnr: str
    charge: str
    mass: Optional[str] = None
    typeB: Optional[str] = None
    chargeB: Optional[str] = None
    massB: Optional[str] = None
    bound_to_nrs: list[str] = field(default_factory=list, compare=False)
    is_radical: bool = field(default=False, compare=False)
    comment: Optional[str] = field(default=None, compare=False)

    @classmethod
    def from_top_line(cls, l: list[str]):
        return cls(
            nr=l[0],
            type=l[1],
            resnr=l[2],
            residue=l[3],
            atom=l[4],
            cgnr=l[5],
            charge=l[6],
            mass=field_or_none(l, 7),
            typeB=field_or_none(l, 8),
            chargeB=field_or_none(l, 9),
            massB=field_or_none(l, 10),
        )


@dataclass()
class PositionRestraint:
    """Information about one position restraint.

    A class containing information as in the position_restraints section of the topology.

    From gromacs topology:
    ; ai   funct    fc(x,y,z)
    """

    ai: str
    funct: str
    fc: tuple[str, str, str]
    condition: Optional[str] = None

    @classmethod
    def from_top_line(cls, l: list[str], condition=None):
        return cls(
            ai=l[0],
            funct=l[1],
            fc=(l[2], l[3], l[4]),
            condition=condition,
        )


@dataclass()
class DihedralRestraint:
    """Information about one dihedral restraint.

    A class containing information as in the dihedral_restraints section of the topology.

    From gromacs topology:
    ; ai   aj    ak    al  type  phi  dphi  fc
    """

    ai: str
    aj: str
    ak: str
    al: str
    type: str
    phi: str
    dphi: str
    fc: str

    @classmethod
    def from_top_line(cls, l: list[str]):
        return cls(
            ai=l[0],
            aj=l[1],
            ak=l[2],
            al=l[3],
            type=l[4],
            phi=l[5],
            dphi=l[6],
            fc=l[7],
        )


@dataclass()
class Settle:
    """Information about one settles

    A class containing atom information as in the settle section of the topology.

    From gromacs topology:
    ; nr funct doh dhh
    """

    nr: str
    funct: str
    doh: str
    dhh: str

    @classmethod
    def from_top_line(cls, l: list[str]):
        return cls(
            nr=l[0],
            funct=l[1],
            doh=l[2],
            dhh=l[3],
        )


@dataclass()
class Exclusion:
    """Information about one exclusion

    A class containing atom information as in the exclusions section of the topology.

    It's unlikey we need this many atomnumbers in a single exclusion, but just in case.
    Because the gromacs manuals just says

    > Each line should start with one atom index, followed by one or more atom indices.
    > All non-bonded interactions between the first atom and the other atoms will be excluded.
    > -- https://manual.gromacs.org/current/reference-manual/topologies/molecule-definition.html#exclusions

    From gromacs topology:
    ; ai aj ak al am an ao ap
    """

    ai: str
    aj: Optional[str] = None
    ak: Optional[str] = None
    al: Optional[str] = None
    am: Optional[str] = None
    an: Optional[str] = None
    ao: Optional[str] = None
    ap: Optional[str] = None
    comment: Optional[str] = field(default=None, compare=False)

    @classmethod
    def from_top_line(cls, l: list[str]):
        return cls(
            ai=l[0],
            aj=field_or_none(l, 1),
            ak=field_or_none(l, 2),
            al=field_or_none(l, 3),
            am=field_or_none(l, 4),
            an=field_or_none(l, 5),
            ao=field_or_none(l, 6),
            ap=field_or_none(l, 7),
        )

    def key(self):
        return tuple(
            field
            for field in [self.ai, self.aj, self.ak, self.al, self.am, self.an, self.ao]
            if field is not None
        )


@dataclass()
class AtomType:
    """Information about one atom type

    A class containing atom type information as in the atomtypes section of the forcefield.

    From gromacs version of the amber* ff:
    ; name at.num mass charge ptype sigma epsilon
    """

    type: str
    id: str
    id_sym: str
    at_num: str
    mass: str
    charge: str
    ptype: str
    sigma: str
    epsilon: str

    @classmethod
    def from_top_line(cls, l: list[str]):
        length = len(l)
        has_at_num = length >= 7
        if has_at_num:
            offset = 1
        else:
            offset = 0
        at_num = l[1] if has_at_num else ""
        return cls(
            type=l[0],
            id=l[0],
            id_sym=l[0],
            at_num=at_num,
            mass=l[1 + offset],
            charge=l[2 + offset],
            ptype=l[3 + offset],
            sigma=l[4 + offset],
            epsilon=l[5 + offset],
        )


@dataclass()
class Bond:
    """Information about one bond

    A class containing bond information as in the bonds section of the topology.

    From gromacs topology:
    ; ai aj funct c0 c1 c2 c3 c4 c5
    With ai < aj
    """

    ai: str
    aj: str
    funct: str
    c0: Optional[str] = None
    c1: Optional[str] = None
    c2: Optional[str] = None
    c3: Optional[str] = None
    c4: Optional[str] = None
    c5: Optional[str] = None
    comment: Optional[str] = field(default=None, compare=False)

    @classmethod
    def from_top_line(cls, l: list[str]):
        funct = field_or_none(l, 2)
        if funct is None:
            funct = FFFUNC["harmonic_bond"]
        return cls(
            ai=l[0],
            aj=l[1],
            funct=funct,
            c0=field_or_none(l, 3),
            c1=field_or_none(l, 4),
            c2=field_or_none(l, 5),
            c3=field_or_none(l, 6),
            c4=field_or_none(l, 7),
            c5=field_or_none(l, 8),
        )


@dataclass()
class BondType:
    """Information about one bondtype

    A class containing bond information as in the bondtype section of the forcefield.

    From gromacs version of the amber* ff:
    ; i j func b0 kb
    Where i and j are atomtypes
    """

    i: str
    j: str
    id: str
    id_sym: str
    funct: str
    c0: Optional[str] = None
    c1: Optional[str] = None
    c2: Optional[str] = None
    c3: Optional[str] = None

    @classmethod
    def from_top_line(cls, l: list[str]):
        funct = field_or_none(l, 2)
        if funct is None:
            funct = FFFUNC["harmonic_bond"]
        return cls(
            i=l[0],
            j=l[1],
            id="---".join(l[:2]),
            id_sym="---".join(reversed(l[:2])),
            funct=funct,
            c0=field_or_none(l, 3),
            c1=field_or_none(l, 4),
            c2=field_or_none(l, 5),
            c3=field_or_none(l, 6),
        )


@dataclass()
class NonbondParamType:
    """Information about one nonbonded parameterize

    typical in coarse grained models.
    A class containing nonbonded information as in the nonbond_params section of the forcefield.

    From gromacs:
    ; Lennard jones between beads
    ; i j funda sigma(nm) epsilon (kmol/mol)
    Where i and j are atomtypes
    """

    i: str
    j: str
    id: str
    id_sym: str
    funct: str
    c0: Optional[str] = None
    c1: Optional[str] = None
    c2: Optional[str] = None

    @classmethod
    def from_top_line(cls, l: list[str]):
        return cls(
            i=l[0],
            j=l[1],
            id="---".join(l[:2]),
            id_sym="---".join(reversed(l[:2])),
            funct=l[2],
            c0=field_or_none(l, 3),
            c1=field_or_none(l, 4),
            c2=field_or_none(l, 5),
        )


@dataclass()
class Pair:
    """Information about one pair

    A class containing pair information as in the pair section of the topology.

    From gromacs topology:
    ; ai aj funct c0 c1 c2 c3 ; comment
    """

    ai: str
    aj: str
    funct: str
    c0: Optional[str] = None
    c1: Optional[str] = None
    c2: Optional[str] = None
    c3: Optional[str] = None
    comment: Optional[str] = field(default=None, compare=False)

    @classmethod
    def from_top_line(cls, l: list[str]):
        return cls(
            ai=l[0],
            aj=l[1],
            funct=l[2],
            c0=field_or_none(l, 3),
            c1=field_or_none(l, 4),
            c2=field_or_none(l, 5),
            c3=field_or_none(l, 6),
        )


@dataclass()
class Angle:
    """Information about one angle

    A class containing angle information as in the angles section of the topology.

    From gromacs topology:
    ; ai aj ak funct c0 c1 c2 c3
    With aj < ai < ak
    """

    ai: str
    aj: str
    ak: str
    funct: str
    c0: Optional[str] = None
    c1: Optional[str] = None
    c2: Optional[str] = None
    c3: Optional[str] = None
    comment: Optional[str] = field(default=None, compare=False)

    @classmethod
    def from_top_line(cls, l: list[str]):
        return cls(
            ai=l[0],
            aj=l[1],
            ak=l[2],
            funct=l[3],
            c0=field_or_none(l, 4),
            c1=field_or_none(l, 5),
            c2=field_or_none(l, 6),
            c3=field_or_none(l, 7),
        )


@dataclass()
class AngleType:
    """Information about one angletype

    A class containing angle type information as in the angletypes section of the forcefield.

    From gromacs version of the amber* ff:
    ; i j k func th0 cth
    """

    i: str
    j: str
    k: str
    id: str
    id_sym: str
    funct: str
    c0: Optional[str] = None
    c1: Optional[str] = None
    c2: Optional[str] = None
    c3: Optional[str] = None

    @classmethod
    def from_top_line(cls, l: list[str]):
        return cls(
            i=l[0],
            j=l[1],
            k=l[2],
            id="---".join(l[:3]),
            id_sym="---".join(reversed(l[:3])),
            funct=l[3],
            c0=field_or_none(l, 4),
            c1=field_or_none(l, 5),
            c2=field_or_none(l, 6),
            c3=field_or_none(l, 7),
        )


@dataclass()
class Dihedral:
    """Information about one proper or improper dihedral

    A class containing dihedral information as in the dihedrals section of the topology.
    Improper dihedrals have funct 4.
    Proper dihedrals have funct != 4, mostly funct 9.

    Note that proper dihedrals of type 9 can be defined multiple times, for different
    periodicities. This is why would-be parameter c2 is called periodicity.

    From gromacs topology:
    ; ai aj ak al funct c0 c1 c2 c3 c4 c5
    For proper dihedrals (funct 9): aj < ak
    """

    ai: str
    aj: str
    ak: str
    al: str
    funct: str
    c0: Optional[str] = None
    c1: Optional[str] = None
    periodicity: str = ""
    c3: Optional[str] = None
    c4: Optional[str] = None
    c5: Optional[str] = None
    comment: Optional[str] = field(default=None, compare=False)

    @classmethod
    def from_top_line(cls, l: list[str]):
        periodicity = field_or_none(l, 7)
        if periodicity is None:
            periodicity = ""
        return cls(
            ai=l[0],
            aj=l[1],
            ak=l[2],
            al=l[3],
            funct=l[4],
            c0=field_or_none(l, 5),
            c1=field_or_none(l, 6),
            periodicity=periodicity,
            c3=field_or_none(l, 8),
            c4=field_or_none(l, 9),
            c5=field_or_none(l, 10),
        )


@dataclass
class MultipleDihedrals:
    """
    Multiple ``Dihedral``s with the same ai, aj, ak, al
    but different periodicities.
    funct should always be "9" when the length of dihedrals is > 1.
    The key of the dihedrals dict is the periodicity (c2).
    """

    ai: str
    aj: str
    ak: str
    al: str
    funct: str
    dihedrals: dict[str, Dihedral]


@dataclass()
class DihedralType:
    """Information about one dihedraltype

    A class containing dihedral type information as in the dihedraltypes
    section of the forcefield.
    Improper dihedrals have funct 4. Proper dihedrals have funct 9.

    Note that proper dihedrals of type 9 can be defined multiple times, for different
    periodicities. This is why would-be parameter c2 is called periodicity and part of
    the `id`.

    From gromacs version of the amber* ff:
    ; i j k l func phase kd pn
    """

    i: str
    j: str
    k: str
    l: str
    id: str
    id_sym: str
    funct: str
    c0: str
    c1: str
    periodicity: str
    c3: Optional[str] = None
    c4: Optional[str] = None
    c5: Optional[str] = None

    @classmethod
    def from_top_line(cls, l: list[str]):
        periodicity = field_or_none(l, 7)
        if periodicity is None:
            periodicity = "2"
        return cls(
            i=l[0],
            j=l[1],
            k=l[2],
            l=l[3],
            id="---".join(l[:4]) + ":::" + periodicity,
            id_sym="---".join(reversed(l[:4])) + ":::" + periodicity,
            funct=l[4],
            c0=l[5],
            c1=l[6],
            periodicity=periodicity,
            c3=field_or_none(l, 8),
            c4=field_or_none(l, 9),
            c5=field_or_none(l, 10),
        )


@dataclass()
class MultipleDihedralTypes:
    """
    Multiple ``DihedralTypes``s with the same ai, aj, ak, al
    but different periodicities.
    funct should always be "9" when the length of dihedrals is > 1.
    The key of the dihedral_types dict is the periodicity (c2).
    """

    ai: str
    aj: str
    ak: str
    al: str
    funct: str
    dihedral_types: dict[str, DihedralType]


@dataclass()
class ResidueAtomSpec:
    """Information about one atom in a residue

    ; name type charge chargegroup
    """

    name: str
    type: str
    charge: str
    cgrp: str

    @classmethod
    def from_top_line(cls, l: list[str]):
        return cls(name=l[0], type=l[1], charge=l[2], cgrp=l[3])


@dataclass()
class ResidueBondSpec:
    """Information about one bond in a residue

    ; atom1 atom2 b0 kb
    """

    atom1: str
    atom2: str
    b0: Optional[str] = None
    kb: Optional[str] = None

    @classmethod
    def from_top_line(cls, l: list[str]):
        return cls(
            atom1=l[0], atom2=l[1], b0=field_or_none(l, 2), kb=field_or_none(l, 3)
        )


@dataclass()
class ResidueImproperSpec:
    """Information about one improper dihedral in a residue

    ; atom1 atom2 atom3 atom4 c0(q0) c1(cp) c2(mult)
    """

    atom1: str
    atom2: str
    atom3: str
    atom4: str
    c0: Optional[str]
    c1: Optional[str]
    c2: Optional[str]

    @classmethod
    def from_top_line(cls, l: list[str]):
        return cls(
            atom1=l[0],
            atom2=l[1],
            atom3=l[2],
            atom4=l[3],
            c0=field_or_none(l, 4),
            c1=field_or_none(l, 5),
            c2=field_or_none(l, 6),
        )

    def reversed(self):
        """Return a reversed version of the improper dihedral"""
        return ResidueImproperSpec(
            atom1=self.atom4,
            atom2=self.atom3,
            atom3=self.atom2,
            atom4=self.atom1,
            c0=self.c0,
            c1=self.c1,
            c2=self.c2,
        )


@dataclass()
class ResidueProperSpec:
    """Information about one proper dihedral in a residue

    ; atom1 atom2 atom3 atom4 c0(q0) c1(cq) c2
    """

    atom1: str
    atom2: str
    atom3: str
    atom4: str
    c0: Optional[str]
    c1: Optional[str]
    c2: Optional[str]

    @classmethod
    def from_top_line(cls, l: list[str]):
        return cls(
            atom1=l[0],
            atom2=l[1],
            atom3=l[2],
            atom4=l[3],
            c0=field_or_none(l, 4),
            c1=field_or_none(l, 5),
            c2=field_or_none(l, 6),
        )


@dataclass()
class ResidueType:
    """Information about one residuetype from aminoacids.rtp"""

    residue: str
    atoms: dict[str, ResidueAtomSpec]
    bonds: dict[tuple[str, str], ResidueBondSpec]
    proper_dihedrals: dict[tuple[str, str, str, str], ResidueProperSpec]
    improper_dihedrals: dict[tuple[str, str, str, str], ResidueImproperSpec]

    @classmethod
    def from_section(cls, residue, d: dict[str, dict[str, list[list[str]]]]):
        atoms = {}
        bonds = {}
        propers = {}
        impropers = {}
        if ls := d.get("atoms"):
            for l in ls["content"]:
                atom = ResidueAtomSpec.from_top_line(l)
                atoms[atom.name] = atom
        if ls := d.get("bonds"):
            for l in ls["content"]:
                bond = ResidueBondSpec.from_top_line(l)
                bonds[(bond.atom1, bond.atom2)] = bond
        if ls := d.get("dihedrals"):
            for l in ls["content"]:
                proper = ResidueProperSpec.from_top_line(l)
                propers[(proper.atom1, proper.atom2, proper.atom3, proper.atom4)] = (
                    proper
                )
        if ls := d.get("impropers"):
            for l in ls["content"]:
                improper = ResidueImproperSpec.from_top_line(l)
                impropers[
                    (improper.atom1, improper.atom2, improper.atom3, improper.atom4)
                ] = improper

        return cls(residue, atoms, bonds, propers, impropers)


AtomId = str
BondId = tuple[str, str]
AngleId = tuple[str, str, str]
DihedralId = tuple[str, str, str, str]
ProperDihedralId = tuple[
    str, str, str, str, str
]  # the complete proper dihedral id also includes the periodicity
ImproperDihedralId = tuple[str, str, str, str]
Atomic = Union[Atom, Bond, Pair, Angle, Dihedral]
AtomicType = Union[AtomType, BondType, AngleType, DihedralType]
AtomicTypes = Union[
    dict[AtomId, AtomType],
    dict[BondId, BondType],
    dict[AngleId, AngleType],
    dict[ProperDihedralId, DihedralType],
    dict[ImproperDihedralId, DihedralType],
]
InteractionIds = Union[BondId, AngleId, ProperDihedralId, ImproperDihedralId]
Interaction = Union[Bond, Pair, Angle, Dihedral]
Interactions = Union[
    dict[BondId, Bond],
    dict[AngleId, Angle],
    dict[ProperDihedralId, Dihedral],
    dict[ImproperDihedralId, Dihedral],
]
InteractionType = Union[BondType, AngleType, DihedralType]
InteractionTypes = Union[
    dict[BondId, BondType],
    dict[AngleId, AngleType],
    dict[ProperDihedralId, DihedralType],
    dict[ImproperDihedralId, DihedralType],
]



# This file is part of the gmx-top4py project.
#
# The gmx-top4py project is based on or includes code from:
#    kimmdy (https://github.com/graeter-group/kimmdy/tree/main)
#    Copyright (C) graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# Modifications and additional code:
#    Copyright (C) 2025 graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import logging
import textwrap
from copy import copy, deepcopy
from pathlib import Path
from typing import Callable, Optional, Union

# from kimmdy.recipe import RecipeStep, Bind, Break

from gmx_top4py.constants import (
    ATOM_ID_FIELDS,
    ATOMTYPE_BONDORDER_FLAT,
    FFFUNC,
    RESNR_ID_FIELDS,
)
from gmx_top4py.parsing import TopologyDict, empty_section, read_top, write_top
from gmx_top4py.parameterizing import BasicParameterizer, Parameterizer

from gmx_top4py.topology.atomic import (
    Angle,
    Atom,
    Bond,
    Dihedral,
    DihedralRestraint,
    Exclusion,
    ImproperDihedralId,
    MoleculeTypeHeader,
    MultipleDihedrals,
    Pair,
    PositionRestraint,
    ResidueImproperSpec,
    ResidueType,
    Settle,
)
from gmx_top4py.topology.ff import FF
from gmx_top4py.topology.utils import (
    attributes_to_list,
    get_moleculetype_atomics,
    get_moleculetype_header,
    get_residue_by_bonding,
    # get_residue_fragments,
    get_top_section,
    increment_field,
    is_not_solvent_or_ion,
    set_top_section,
)
from gmx_top4py.utils import TopologyAtomAddress

logger = logging.getLogger("gmx_top4py.topology")


class MoleculeType:
    """One moleculetype in the topology

    Attributes
    ----------
    atoms : dict[str, Atom]
    bonds : dict[tuple[str, str], Bond]
    pairs : dict[tuple[str, str], Pair]
    angles : dict[tuple[str, str, str], Angle]
    proper_dihedrals : dict[tuple[str, str, str, str], MultipleDihedrals]
    improper_dihedrals : dict[tuple[str, str, str, str], Dihedral]
    position_restraints : dict[str, PositionRestraint]
    dihedral_restraints : dict[tuple[str, str, str, str], DihedralRestraint]
    radicals : dict[str, Atom]
        dict mapping atom indices to atom objects for storing all radical atoms

    """

    def __init__(
        self, header: MoleculeTypeHeader, atomics: dict, radicals: Optional[str] = None
    ) -> None:
        self.name = header.name
        self.nrexcl = header.nrexcl
        self.atomics = atomics

        logger.debug(f"parsing molecule {self.name}")

        self.atoms: dict[str, Atom] = {}
        self.bonds: dict[tuple[str, str], Bond] = {}
        self.pairs: dict[tuple[str, str], Pair] = {}
        self.angles: dict[tuple[str, str, str], Angle] = {}
        self.proper_dihedrals: dict[tuple[str, str, str, str], MultipleDihedrals] = {}
        self.improper_dihedrals: dict[tuple[str, str, str, str], MultipleDihedrals] = {}
        self.position_restraints: dict[str, PositionRestraint] = {}
        self.dihedral_restraints: dict[tuple[str, str, str, str], DihedralRestraint] = (
            {}
        )
        self.settles: dict[str, Settle] = {}
        self.exclusions: dict[tuple, Exclusion] = {}

        self._parse_atoms()
        self._parse_bonds()
        self._parse_pairs()
        self._parse_angles()
        self._parse_dihedrals()
        self._parse_restraints()
        self._parse_settles()
        self._parse_exclusions()
        self._initialize_graph()

        # must be after self._parse_atoms
        self.radicals: dict[str, Atom] = {}
        if radicals is not None:
            # if self.name == "Reactive" and len(radicals) > 0:
            if len(radicals) > 0:
                logger.debug(
                    f"Using 'radicals' section from config file with entries: '{radicals}'."
                )
                for radical in radicals.strip().split(sep=" "):
                    atom = self.atoms.get(radical)
                    if atom:
                        self.radicals[radical] = atom
                        atom.is_radical = True
                    else:
                        logger.debug(
                            f"Atom nr {radical} from 'radicals' section in "
                            "config file not in topology. Ignoring this entry."
                        )
        else:
            self.find_radicals()

    def __str__(self) -> str:
        return textwrap.dedent(
            f"""\
        Moleculetype {self.name} with:
        {len(self.atoms)} atoms,
        {len(self.bonds)} bonds,
        {len(self.angles)} angles,
        {len(self.pairs)} pairs,
        {len(self.settles)} settles,
        {len(self.exclusions)} exclusions,
        {len(self.proper_dihedrals)} proper dihedrals
        {len(self.improper_dihedrals)} improper dihedrals
        {len(self.position_restraints)} position restraints
        {len(self.dihedral_restraints)} dihedral restraints
        """
        )

    def __repr__(self) -> str:
        return f"Molecule({(self.name, self.nrexcl)}, {self.atomics})"

    def _repr_pretty_(self, p, _):
        """A __repr__ for ipython.

        This whill be used if just the name of the object is entered in the ipython shell
        or a jupyter notebook.

        p is an instance of
        [IPython.lib.pretty.PrettyPrinter](https://ipython.org/ipython-doc/3/api/generated/IPython.lib.pretty.html)
        """
        p.text(str(self))

    def _parse_atoms(self):
        """Parse atoms from topology dictionary."""
        ls = self.atomics.get("atoms")
        if ls is None:
            raise ValueError(f"No atoms found in topology for molecule {self.name}.")
        for l in ls:
            atom = Atom.from_top_line(l)
            self.atoms[atom.nr] = atom

    def _parse_bonds(self):
        """Parse bond from topology dictionary."""
        ls = self.atomics.get("bonds")
        if ls is None:
            return
        for l in ls:
            bond = Bond.from_top_line(l)
            self.bonds[(bond.ai, bond.aj)] = bond

    def _parse_pairs(self):
        """Parse pairs from topology dictionary."""
        ls = self.atomics.get("pairs")
        if ls is None:
            return
        for l in ls:
            pair = Pair.from_top_line(l)
            self.pairs[(pair.ai, pair.aj)] = pair

    def _parse_angles(self):
        """Parse angles from topology dictionary."""
        ls = self.atomics.get("angles")
        if ls is None:
            return
        for l in ls:
            angle = Angle.from_top_line(l)
            self.angles[(angle.ai, angle.aj, angle.ak)] = angle

    def _parse_dihedrals(self):
        """Parse improper and proper dihedrals from topology dictionary."""
        ls = self.atomics.get("dihedrals")
        if ls is None:
            return
        for l in ls:
            dihedral = Dihedral.from_top_line(l)
            key = (dihedral.ai, dihedral.aj, dihedral.ak, dihedral.al)
            if dihedral.funct == FFFUNC["mult_improper_dihedral"]:
                if self.improper_dihedrals.get(key) is None:
                    self.improper_dihedrals[key] = MultipleDihedrals(
                        *key, dihedral.funct, dihedrals={}
                    )
                self.improper_dihedrals[key].dihedrals[dihedral.periodicity] = dihedral
            else:
                if self.proper_dihedrals.get(key) is None:
                    self.proper_dihedrals[key] = MultipleDihedrals(
                        *key, dihedral.funct, dihedrals={}
                    )
                self.proper_dihedrals[key].dihedrals[dihedral.periodicity] = dihedral

    def _parse_restraints(self):
        """Parse restraints from topology dictionary."""
        ls = self.atomics.get("position_restraints")
        if not ls:
            ls = []
        condition = None
        for l in ls:
            restraint = PositionRestraint.from_top_line(l, condition=condition)
            self.position_restraints[restraint.ai] = restraint
        ls = self.atomics.get("dihedral_restraints")
        if not ls:
            ls = []
        for l in ls:
            restraint = DihedralRestraint.from_top_line(l)
            self.dihedral_restraints[
                (restraint.ai, restraint.aj, restraint.ak, restraint.al)
            ] = restraint

    def _parse_settles(self):
        """Parse settles from topology dictionary."""
        ls = self.atomics.get("settles")
        if not ls:
            return
        for l in ls:
            settle = Settle.from_top_line(l)
            self.settles[settle.nr] = settle

    def _parse_exclusions(self):
        """Parse exclusions from topology dictionary."""
        ls = self.atomics.get("exclusions")
        if not ls:
            return
        for _, l in enumerate(ls):
            exclusion = Exclusion.from_top_line(l)
            self.exclusions[exclusion.key()] = exclusion

    def _initialize_graph(self):
        """Add a list of atom nrs bound to an atom to each atom."""
        for bond in self.bonds.values():
            i = bond.ai
            j = bond.aj
            self.atoms[i].bound_to_nrs.append(j)
            self.atoms[j].bound_to_nrs.append(i)

    def find_radicals(self):
        """Find atoms that are radicals and update self.radicals.

        Iterate over all atoms and designate them as radicals if they have
        fewer bounds than their natural bond order.
        """
        logger.debug(
            "Trying to infer radical status based on AMBER(!!) atomtype bond "
            f"order in molecule {self.name}."
        )
        self.radicals = {}
        error_atoms = set()
        for atom in self.atoms.values():
            bo = ATOMTYPE_BONDORDER_FLAT.get(atom.type)
            if bo is None:
                error_atoms.add(atom.type)
            if bo and bo > len(atom.bound_to_nrs):
                self.radicals[atom.nr] = atom
                atom.is_radical = True
            else:
                atom.is_radical = False
        if len(error_atoms) > 0:
            logger.warning(
                "Some atomtypes not found in AMBER atomtypes. Cannot infer radical status.\n"
                "Missing atom types:\n"
                f"{error_atoms}"
            )
        return None

    def _update_atomics_dict(self):
        self.atomics["atoms"] = [attributes_to_list(x) for x in self.atoms.values()]
        self.atomics["bonds"] = [attributes_to_list(x) for x in self.bonds.values()]
        self.atomics["angles"] = [attributes_to_list(x) for x in self.angles.values()]
        self.atomics["pairs"] = [attributes_to_list(x) for x in self.pairs.values()]
        self.atomics["dihedrals"] = [
            attributes_to_list(x)
            for dihedrals in self.proper_dihedrals.values()
            for x in dihedrals.dihedrals.values()
        ] + [
            attributes_to_list(x)
            for dihedrals in self.improper_dihedrals.values()
            for x in dihedrals.dihedrals.values()
        ]
        self.atomics["position_restraints"] = [
            attributes_to_list(x) for x in self.position_restraints.values()
        ]
        self.atomics["dihedral_restraints"] = [
            attributes_to_list(x) for x in self.dihedral_restraints.values()
        ]
        self.atomics["settles"] = [attributes_to_list(x) for x in self.settles.values()]
        self.atomics["exclusions"] = [
            attributes_to_list(x) for x in self.exclusions.values()
        ]

    def _get_atom_bonds(self, atom_nr: str) -> list[tuple[str, str]]:
        """Get all bonds a particular atom is involved in."""
        ai = atom_nr
        bonds = []
        for aj in self.atoms[ai].bound_to_nrs:
            if int(ai) < int(aj):
                bonds.append((ai, aj))
            else:
                bonds.append((aj, ai))
        return bonds

    def _get_atom_pairs(self, _: str) -> list[tuple[str, str]]:
        raise NotImplementedError(
            "get_atom_pairs is not implemented. Get the pairs as the endpoints of dihedrals instead."
        )

    def _get_atom_angles(self, atom_nr: str) -> list[tuple[str, str, str]]:
        """
        Gets a list of angles that one atom is involved in based on the bond
        information stored in the atom.

        Angles between atoms ai, aj, ak satisfy ai < ak
        """
        return self._get_center_atom_angles(atom_nr) + self._get_margin_atom_angles(
            atom_nr
        )

    def _get_center_atom_angles(self, atom_nr: str) -> list[tuple[str, str, str]]:
        # atom_nr in the middle of an angle
        angles = []
        aj = atom_nr
        partners = self.atoms[aj].bound_to_nrs
        for ai in partners:
            for ak in partners:
                if int(ai) < int(ak):
                    angles.append((ai, aj, ak))
        return angles

    def _get_margin_atom_angles(self, atom_nr: str) -> list[tuple[str, str, str]]:
        # atom_nr at the outer corner of angle
        angles = []
        ai = atom_nr
        for aj in self.atoms[ai].bound_to_nrs:
            for ak in self.atoms[aj].bound_to_nrs:
                if ai == ak:
                    continue
                if int(ai) < int(ak):
                    angles.append((ai, aj, ak))
                else:
                    angles.append((ak, aj, ai))
        return angles

    def _get_atom_proper_dihedrals(
        self, atom_nr: str
    ) -> list[tuple[str, str, str, str]]:
        """
        Gets a list of dihedrals that one atom is involved in based on the bond
        information stored in the atom.
        """
        return self._get_center_atom_dihedrals(
            atom_nr
        ) + self._get_margin_atom_dihedrals(atom_nr)

    def _get_center_atom_dihedrals(
        self, atom_nr: str, improper: bool = False
    ) -> list[tuple[str, str, str, str]]:
        """
        propers are ordered by the middle atoms j,k
        impropers are not ordered
        """
        dihedrals = []
        aj = atom_nr
        partners = self.atoms[aj].bound_to_nrs
        for ai in partners:
            for ak in partners:
                if ai == ak:
                    continue
                for al in self.atoms[ak].bound_to_nrs:
                    if int(aj) < int(ak) or improper:
                        key = (ai, aj, ak, al)
                    else:
                        key = (al, ak, aj, ai)
                    if len(key) == len(set(key)):
                        # one atom can not be in two positions
                        dihedrals.append(key)
        return dihedrals

    def _get_margin_atom_dihedrals(
        self, atom_nr: str, improper: bool = False
    ) -> list[tuple[str, str, str, str]]:
        """
        propers are ordered by the middle atoms j,k
        impropers are not ordered
        """
        dihedrals = []
        ai = atom_nr
        for aj in self.atoms[ai].bound_to_nrs:
            for ak in self.atoms[aj].bound_to_nrs:
                if ai == ak:
                    continue
                for al in self.atoms[ak].bound_to_nrs:
                    if int(aj) < int(ak) or improper:
                        key = (ai, aj, ak, al)
                    else:
                        key = (al, ak, aj, ai)
                    if len(key) == len(set(key)):
                        # one atom can not be in two positions
                        dihedrals.append(key)
        return dihedrals

    def _get_atom_improper_dihedrals(
        self, atom_nr: str, ff: FF
    ) -> dict[ImproperDihedralId, ResidueImproperSpec]:
        """
        <https://manual.gromacs.org/current/reference-manual/functions/bonded-interactions.html#improper-dihedrals>
        """
        atom = self.atoms[atom_nr]
        residuetype = ff.residuetypes.get(atom.residue)
        if not isinstance(residuetype, ResidueType):
            return {}

        residue = get_residue_by_bonding(atom=atom, atoms=self.atoms)
        lowest_atomnr = min(residue.keys(), key=int)
        highest_atomnr = max(residue.keys(), key=int)

        previous_atom = self.atoms.get(str(int(lowest_atomnr) - 1))
        next_atom = self.atoms.get(str(int(highest_atomnr) + 1))
        if previous_atom is not None:
            previous_residue = get_residue_by_bonding(
                atom=previous_atom, atoms=self.atoms
            )
            previous_residuetype = ff.residuetypes.get(previous_atom.residue)
        else:
            previous_residue = {}
            previous_residuetype = None
        if next_atom is not None:
            next_residue = get_residue_by_bonding(atom=next_atom, atoms=self.atoms)
            next_residuetype = ff.residuetypes.get(next_atom.residue)
        else:
            next_residue = {}
            next_residuetype = None

        atomname_to_nr = {atom.atom: atom_nr for atom_nr, atom in residue.items()}
        # residues on aminoacids.rtp specify a dihedral to the next or previous
        # AA with -C and +N as the atomname
        prev_atomname_to_nr = {
            atom.atom: atom_nr for atom_nr, atom in previous_residue.items()
        }
        next_atomname_to_nr = {
            atom.atom: atom_nr for atom_nr, atom in next_residue.items()
        }

        dihedrals: dict[ImproperDihedralId, ResidueImproperSpec] = {}
        if previous_residuetype is not None and atom.atom == "N":
            # the atom is N, so it may have an improper defined
            # in the previous residue, where it would be referred to as +N
            for key, improper in previous_residuetype.improper_dihedrals.items():
                # find the dihedral with +N
                if any(["+" in k for k in key]):
                    nr_key: ImproperDihedralId = tuple(
                        atom.nr if "+" in x else prev_atomname_to_nr.get(x) for x in key
                    )  # pyright: ignore
                    if None in nr_key:
                        m = f"Improper dihedral {key} with {nr_key} not found in residue {atom.residue}."
                        logger.debug(m)
                        continue
                    dihedrals[nr_key] = improper

        if next_residuetype is not None and atom.atom == "C":
            # the atom is C, so it may have an improper defined
            # in the previous residue, where it would be referred to as -C
            for key, improper in next_residuetype.improper_dihedrals.items():
                # find the dihedral with -C
                if any(["-" in k for k in key]):
                    nr_key = tuple(
                        atom.nr if "-" in x else next_atomname_to_nr.get(x) for x in key
                    )  # pyright: ignore
                    if None in nr_key:
                        m = f"Improper dihedral {key} with {nr_key} not found in residue {atom.residue}."
                        logger.debug(m)
                        continue
                    dihedrals[nr_key] = improper

        def get_number(name):
            """Get the atom number for an atom name from a dihedral specification"""
            if "+" in name:
                # + refers to the next residue
                return next_atomname_to_nr.get(name.replace("+", ""), None)
            elif "-" in name:
                # - refers to the previous residue
                return prev_atomname_to_nr.get(name.replace("-", ""), None)
            else:
                # regular atom name
                return atomname_to_nr.get(name, None)

        # regularly defined improper dihedrals of the same residue
        for key, improper in residuetype.improper_dihedrals.items():
            nr_key = tuple(get_number(x) for x in key)  # pyright: ignore
            if atom_nr not in nr_key:
                continue
            if None in nr_key:
                m = f"Improper dihedral {key} {nr_key} not found in residue {atom.residue}."
                logger.debug(m)
                continue
            dihedrals[nr_key] = improper
        # gromacs sorts the keys such that the first atom is the lowest numbered
        new_dihedrals: dict[ImproperDihedralId, ResidueImproperSpec] = {}
        for key, improper in dihedrals.items():
            if int(key[0]) > int(key[3]):
                key = (key[3], key[2], key[1], key[0])
                new_dihedrals[key] = improper.reversed()
            else:
                new_dihedrals[key] = improper

        return new_dihedrals

    def _regenerate_topology_from_bound_to(self, ff):
        # clear all bonds, angles, dihedrals
        self.bonds.clear()
        self.angles.clear()
        self.proper_dihedrals.clear()
        self.proper_dihedrals.clear()
        self.improper_dihedrals.clear()

        # bonds
        keys = []
        for atom in self.atoms.values():
            keys = self._get_atom_bonds(atom.nr)
            for key in keys:
                self.bonds[key] = Bond(key[0], key[1], FFFUNC["harmonic_bond"])

        # angles
        for atom in self.atoms.values():
            keys = self._get_atom_angles(atom.nr)
            for key in keys:
                self.angles[key] = Angle(
                    key[0], key[1], key[2], FFFUNC["harmonic_angle"]
                )

        # dihedrals and pass
        for atom in self.atoms.values():
            keys = self._get_atom_proper_dihedrals(atom.nr)
            for key in keys:
                self.proper_dihedrals[key] = MultipleDihedrals(
                    *key,
                    FFFUNC["mult_proper_dihedral"],
                    dihedrals={"": Dihedral(*key, FFFUNC["mult_proper_dihedral"])},
                )
                pairkey: tuple[str, str] = tuple(
                    str(x) for x in sorted([key[0], key[3]], key=int)
                )  # type: ignore
                if self.pairs.get(pairkey) is None:
                    self.pairs[pairkey] = Pair(pairkey[0], pairkey[1], FFFUNC["pair"])

        for atom in self.atoms.values():
            impropers = self._get_atom_improper_dihedrals(atom.nr, ff)
            for key, improper in impropers.items():
                periodicity = ""
                if improper.c2 is not None:
                    periodicity = improper.c2
                self.improper_dihedrals[key] = MultipleDihedrals(
                    *key,
                    FFFUNC["mult_improper_dihedral"],
                    dihedrals={
                        "": Dihedral(
                            *key,
                            FFFUNC["mult_improper_dihedral"],
                            c0=improper.c0,
                            c1=improper.c1,
                            periodicity=periodicity,
                        )
                    },
                )

    def reindex_atomnrs(self) -> dict[str, str]:
        """Reindex atom numbers in topology.

        Starts at index 1.
        This also updates the numbers for bonds, angles, dihedrals and pairs.
        Returns a dict, mapping of old atom number strings to new ones
        """

        update_map = {
            atom_nr: str(i + 1)
            for i, atom_nr in enumerate(sorted(list(self.atoms.keys()), key=int))
        }

        new_atoms = {}
        for atom in self.atoms.values():
            atom.nr = update_map[atom.nr]
            bound_to = []
            for nr in atom.bound_to_nrs:
                new_nr = update_map.get(nr)
                if new_nr is not None:
                    bound_to.append(new_nr)
            atom.bound_to_nrs = bound_to
            new_atoms[atom.nr] = atom
        self.atoms = new_atoms

        new_bonds = {}
        for bond in self.bonds.values():
            ai = update_map.get(bond.ai)
            aj = update_map.get(bond.aj)
            # drop bonds to a deleted atom
            if ai is None or aj is None:
                continue
            bond.ai = ai
            bond.aj = aj
            new_bonds[(ai, aj)] = bond
        self.bonds = new_bonds

        new_angles = {}
        for angle in self.angles.values():
            ai = update_map.get(angle.ai)
            aj = update_map.get(angle.aj)
            ak = update_map.get(angle.ak)
            # drop angles to a deleted atom
            if None in (ai, aj, ak):
                continue
            # pyright does not grok `None in ...`
            angle.ai = ai  # type: ignore
            angle.aj = aj  # type: ignore
            angle.ak = ak  # type: ignore
            new_angles[(angle.ai, angle.aj, angle.ak)] = angle
        self.angles = new_angles

        new_pairs = {}
        new_multiple_dihedrals = {}
        for dihedrals in self.proper_dihedrals.values():
            new_dihedrals = {}
            ai = update_map.get(dihedrals.ai)
            aj = update_map.get(dihedrals.aj)
            ak = update_map.get(dihedrals.ak)
            al = update_map.get(dihedrals.al)
            # drop dihedrals to a deleted atom
            if None in (ai, aj, ak, al):
                continue

            # do pairs before the dihedrals are updated
            pair = self.pairs.pop((dihedrals.ai, dihedrals.al), None)
            if pair is not None:
                pair_ai = update_map.get(pair.ai)
                pair_aj = update_map.get(pair.aj)

                if None not in (pair_ai, pair_aj):
                    pair.ai = pair_ai  # type: ignore (pyright bug)
                    pair.aj = pair_aj  # type: ignore
                    new_pairs[(pair_ai, pair_aj)] = pair

            dihedrals.ai = ai  # type: ignore
            dihedrals.aj = aj  # type: ignore
            dihedrals.ak = ak  # type: ignore
            dihedrals.al = al  # type: ignore

            for dihedral in dihedrals.dihedrals.values():
                dihedral.ai = ai  # type: ignore
                dihedral.aj = aj  # type: ignore
                dihedral.ak = ak  # type: ignore
                dihedral.al = al  # type: ignore
                new_dihedrals[dihedral.periodicity] = dihedral
            dihedrals.dihedrals = new_dihedrals

            new_multiple_dihedrals[
                (
                    dihedrals.ai,
                    dihedrals.aj,
                    dihedrals.ak,
                    dihedrals.al,
                )
            ] = dihedrals

        self.proper_dihedrals = new_multiple_dihedrals
        self.pairs = new_pairs

        new_multiple_dihedrals = {}
        for dihedrals in self.improper_dihedrals.values():
            new_dihedrals = {}
            ai = update_map.get(dihedrals.ai)
            aj = update_map.get(dihedrals.aj)
            ak = update_map.get(dihedrals.ak)
            al = update_map.get(dihedrals.al)
            # drop dihedrals to a deleted atom
            if None in (ai, aj, ak, al):
                continue
            dihedrals.ai = ai  # type: ignore
            dihedrals.aj = aj  # type: ignore
            dihedrals.ak = ak  # type: ignore
            dihedrals.al = al  # type: ignore

            for dihedral in dihedrals.dihedrals.values():
                dihedral.ai = ai  # type: ignore
                dihedral.aj = aj  # type: ignore
                dihedral.ak = ak  # type: ignore
                dihedral.al = al  # type: ignore
                new_dihedrals[dihedral.periodicity] = dihedral
            dihedrals.dihedrals = new_dihedrals

            new_multiple_dihedrals[(ai, aj, ak, al)] = dihedrals
        self.improper_dihedrals = new_multiple_dihedrals

        return update_map


class Topology:
    """Smart container for parsed topology data.

    A topology keeps track of connections when bonds are broken or formed.
    Reparametrization is triggerd automatically if `to_dict` is called
    after bonds have changed.

    Also see <https://manual.gromacs.org/current/reference-manual/topologies/topology-file-formats.html#topology-file>

    Assumptions:

    - the topology of interest (the Reactive moleculetype) consists of the first moleculetypes (non-solvent).

    Parameters
    ----------
    top
        A dictionary containing the parsed topology data, produced by
        [](`gmx_top4py.parsing.read_top`)
    parametrizer
        The parametrizer to use when reparametrizing the topology.
    is_selected_moleculetype_f
        A function that takes a moleculetype name and returns True if the moleculetype is selected.
        Use this function to mark one or more moleculetypes.
        If an selected moleculetype has multiple copies, this multiplicity of the moleculetype will be made explicit.
        If multiple different moleculetypes are selected, they will be merged into one moleculetype.
        By default, all non-solvent and non-ion moleculetypes are selected.
    radicals
        A string of atom numbers that are radicals.
    residuetypes_path
        Path to the residue types file.
    nrexcl
        Overwrite nrexcl value for the selected moleculetype. Otherwise takes the nrexcl of the first selected moleculetype.
    needs_parameterization
        Does the topology currently need to be re-parameterized due to changes?
    parameterization_focus_ids
        list of atoms ids around which the parameterization happens (to avoid re-parameterizing the whole selected moleculetype)
    """

    @classmethod
    def from_path(cls, path: str | Path, ffdir: str | Path | None = None, **kwargs):
        """Load a Topology object from a file path.

        Simplifies the `Topology(read_top(...), ...)` pattern.

        Parameters
        ----------
        path
            Path to the topology file.
        kwargs
            Additional keyword arguments to pass to the Topology constructor.

        Returns
        -------
        Topology
        """
        if isinstance(path, str):
            path = Path(path)
        if isinstance(ffdir, str):
            ffdir = Path(ffdir)
        return cls(read_top(path=path, ffdir=ffdir), **kwargs)

    def to_path(self, path: str | Path) -> None:
        """Write the topology to a file.

        Parameters
        ----------
        path
            Path to the topology file.
        """
        if isinstance(path, str):
            path = Path(path)
        write_top(self.to_dict(), path)

    def __init__(
        self,
        top: TopologyDict,
        parametrizer: Parameterizer = BasicParameterizer(),
        is_selected_moleculetype_f: Callable[[str], bool] = is_not_solvent_or_ion,
        radicals: Optional[str] = None,
        residuetypes_path: Optional[Path] = None,
        nrexcl: Optional[str] = None,
    ):
        if top == {}:
            raise NotImplementedError(
                "Generating an empty Topology from an empty TopologyDict is not implemented."
            )

        self.top = top
        self.moleculetypes: dict[str, MoleculeType] = {}
        molecules = get_top_section(top, "molecules")
        if molecules is None:
            raise ValueError("molecules not found in top file")
        self.molecules = [(l[0], l[1]) for l in molecules]

        self._parse_ff(top, residuetypes_path)
        self._parse_molecules()
        self.parametrizer = parametrizer
        self._check_is_selected_molecule = is_selected_moleculetype_f

        self.needs_parameterization: bool = False
        self.parameterization_focus_ids: set[str] = set()

        self._merge_moleculetypes(radicals=radicals, nrexcl=nrexcl)
        self._link_atomics()

    def _parse_ff(self, top: TopologyDict, residuetypes_path: Optional[Path] = None):
        self.ff = FF(top, residuetypes_path)

    def _link_atomics(self):
        """link atoms, bonds etc. properties of the selected moleculetype(s) to self.

        Call this any time a large change is made to the selected moleculetype
        that re-assigns a property and not just modifies parts of it in place.
        """
        self.selected_molecule = self.moleculetypes[self.selected_moleculetype]
        self.atoms = self.selected_molecule.atoms
        self.bonds = self.selected_molecule.bonds
        self.angles = self.selected_molecule.angles
        self.exclusions = self.selected_molecule.exclusions
        self.settles = self.selected_molecule.settles
        self.proper_dihedrals = self.selected_molecule.proper_dihedrals
        self.improper_dihedrals = self.selected_molecule.improper_dihedrals
        self.pairs = self.selected_molecule.pairs
        self.position_restraints = self.selected_molecule.position_restraints
        self.dihedral_restraints = self.selected_molecule.dihedral_restraints
        self.radicals = self.selected_molecule.radicals
        self.nrexcl = self.selected_molecule.nrexcl

    def _extract_mergable_molecules(self) -> dict[str, int]:
        """Extract all selected molecules that are to be merged into one moleculetype.
        And replace the selected moleculetypes with a single moleculetype name

        The new moleculetype name is a concatenation of all selected moleculetypes
        with their multiplicity in brackets, e.g. 'MOL1(2x)-MOL2(1x)-MOL3(4x)'.
        """
        selected_molecules = {}
        new_molecules = []
        started_merging = False
        stopped_merging = False

        for m, n in self.molecules:
            if self._check_is_selected_molecule(m):
                if stopped_merging:
                    m = f"""Attempting to merge a moleculetype {m} interspersed with non-merging moleculetypes.
            Please make sure that all moleculetypes to be merged (all non-solvent molecules or ions by default)
            are listed consecutively in the [molecules] section of the topology and in the coordinates file (.gro)
            """
                    logger.error(m)
                    raise ValueError(m)
                started_merging = True
                selected_molecules[m] = int(n)
            else:
                if started_merging:
                    stopped_merging = True
                new_molecules += [(m, n)]

        self.molecules = new_molecules
        self.selected_moleculetype = "-".join(
            [f"{m}({n}x)" for m, n in selected_molecules.items()]
        )
        new_molecules.insert(
            0, (self.selected_moleculetype, "1")
        )  # The selected moleculetype is inserted at the start of the list to comply with previous code behaviour
        logger.debug(
            f"Merging the following molecules into a single moleculetype called {self.selected_moleculetype} and making their multiples explicit:"
        )
        for m, n in selected_molecules.items():
            logger.debug(f"\t{m} {n}")
        return selected_molecules

    def _merge_moleculetypes(
        self, radicals: Optional[str] = None, nrexcl: Optional[str] = None
    ):
        """
        Merge all selected moleculetypes into one moleculetype.
        This also makes multiples explicit.
        Selected molecules have to be in a continuous block in the [molecules] section of the topology
        and the gro file, preferably the start.
        Atom ids (gromacs, 1-based) are updated accordingly and thus correspond directly
        to the atom numbers in the coordinates file (.gro) (ignoring gromacs overflow
        problems in the atomnr column, the correct internal atom nr is always "gro file line number" - 2).
        """
        molecules = self._extract_mergable_molecules()
        first_selected_molecule = list(molecules.keys())[0]
        if nrexcl is None:
            nrexcl = self.moleculetypes[first_selected_molecule].nrexcl
        merged_atomics = {}
        atomnr_offset = 0
        resnr_offset = 0
        for name, n in molecules.items():
            add_atomics = self.moleculetypes[name].atomics
            n_atoms = len(add_atomics["atoms"])
            highest_resnr = int(add_atomics["atoms"][-1][RESNR_ID_FIELDS["atoms"][0]])
            for _ in range(n):
                for section_name, section in add_atomics.items():
                    section = deepcopy(section)
                    atomnr_fields = ATOM_ID_FIELDS.get(section_name, [])
                    resnr_fields = RESNR_ID_FIELDS.get(section_name, [])
                    for line in section:
                        if atomnr_fields is True:
                            # if atomnr_fields is True, then all fields are atomnr fields
                            # NOTE: This is why this is testing for actually being True, not just truthiness!
                            for field, _ in enumerate(line):
                                increment_field(line, field, atomnr_offset)
                            continue
                        for field in atomnr_fields:
                            increment_field(line, field, atomnr_offset)
                        for field in resnr_fields:
                            increment_field(line, field, resnr_offset)
                    if merged_atomics.get(section_name) is None:
                        merged_atomics[section_name] = []
                    merged_atomics[section_name] += section

                atomnr_offset += n_atoms
                resnr_offset += highest_resnr
            # remove now merged moleculetype from topology
            # and the topology dict
            del self.moleculetypes[name]
            del self.top[f"moleculetype_{name}"]

        # add merged moleculetype to topology
        merged_moleculetype = MoleculeType(
            MoleculeTypeHeader(name=self.selected_moleculetype, nrexcl=nrexcl),
            merged_atomics,
            radicals,
        )
        self.moleculetypes[self.selected_moleculetype] = merged_moleculetype
        # update topology dict
        self._update_dict()

    # def validate_bond(self, atm1: Atom, atm2: Atom) -> bool:
    #     """Validates bond consistency between both atoms and top
    #     Returns True if bond exists, False if not.
    #     Raises RuntimeError if bond is not consistent.
    #     """

    #     counter = 0
    #     if (atm1.nr, atm2.nr) in self.bonds.keys():
    #         counter += 1
    #     if atm1.nr in atm2.bound_to_nrs:
    #         counter += 1
    #     if atm2.nr in atm1.bound_to_nrs:
    #         counter += 1
    #     if counter == 3:
    #         return True
    #     elif counter == 0:
    #         return False
    #     raise RuntimeError(
    #         f"Bond between atom {atm1.nr} {atm1.type} and {atm2.nr} {atm2.type} "
    #         "(1-based) is ill-defined!\n"
    #         f"\tatm1.bound_to_nrs {atm1.bound_to_nrs}\n"
    #         f"\tatm2.bound_to_nrs {atm2.bound_to_nrs}\n"
    #     )

    def _parse_molecules(self):
        moleculetypes = [k for k in self.top.keys() if k.startswith("moleculetype")]
        for moleculetype in moleculetypes:
            header = get_moleculetype_header(self.top, moleculetype)
            if header is None:
                logger.warning(f"moleculetype {moleculetype} has no header. Skipping.")
                continue
            atomics = get_moleculetype_atomics(self.top, moleculetype)
            if atomics is None:
                logger.warning(
                    f"moleculetype {moleculetype} has no atoms, bonds, angles etc. Skipping."
                )
                continue
            name = header.name
            # Only search for radicals in the selected moleculetypes during merging
            self.moleculetypes[name] = MoleculeType(header, atomics, radicals="")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Topology):
            return False
        return self.to_dict() == other.to_dict()

    def _update_dict_from_moleculetype_atomics(
        self, moleculetype: MoleculeType, create: bool = False
    ):
        """Set content of the atomics (atoms/bonds/angles etc.) in the a topology dict from of a moleculetype.

        Resolves any `#ifdef` statements by check in the top['define'] dict
        and chooses the 'content' or 'else_content' depending on the result.

        If create is True, a new section is created if it does not exist.
        """
        name = moleculetype.name
        atomics = moleculetype.atomics
        top = self.top
        moleculetype_name = f"moleculetype_{name}"
        section = top.get(moleculetype_name)
        if section is None:
            if create:
                logger.debug(
                    f"topology does not contain {moleculetype_name}. Creating new section."
                )
                section = empty_section()
                section["content"] = [[name, moleculetype.nrexcl]]
                section["subsections"] = {k: empty_section() for k in atomics.keys()}
                top[moleculetype_name] = section
            else:
                logger.warning(
                    f"topology does not contain {moleculetype_name} and create=False. Not creating new section."
                )
                return None

        subsections = section["subsections"]
        section["content"] = [[name, moleculetype.nrexcl]]
        for k in list(subsections.keys()):
            v = subsections[k]
            condition = v.get("condition")
            if condition is not None:
                condition_type = condition.get("type")
                condition_value = condition.get("value")
                if condition_type == "ifdef":
                    if condition_value in top["define"].keys():
                        v["content"] = atomics[k]
                    else:
                        v["else_content"] = atomics[k]
                elif condition_type == "ifndef":
                    if condition_value not in top["define"].keys():
                        v["content"] = atomics[k]
                    else:
                        v["else_content"] = atomics[k]
                else:
                    raise NotImplementedError(
                        f"condition type {condition_type} is not supported"
                    )
            else:
                v["content"] = atomics[k]

    def _update_dict(self):
        """Update the topology dictionary with the current state of the topology."""
        for name, moleculetype in self.moleculetypes.items():
            moleculetype._update_atomics_dict()
            if name == self.selected_moleculetype:
                self._update_dict_from_moleculetype_atomics(moleculetype, create=True)
            else:
                self._update_dict_from_moleculetype_atomics(moleculetype)

        set_top_section(
            self.top,
            "molecules",
            [[name, n] for name, n in self.molecules],
        )

        set_top_section(
            self.top,
            "atomtypes",
            [attributes_to_list(x) for x in self.ff.atomtypes.values()],
        )

        set_top_section(
            self.top,
            "bondtypes",
            [attributes_to_list(x) for x in self.ff.bondtypes.values()],
        )
        if self.ff.nonbond_params != {}:
            set_top_section(
                self.top,
                "nonbond_params",
                [attributes_to_list(x) for x in self.ff.nonbond_params.values()],
            )
        set_top_section(
            self.top,
            "angletypes",
            [attributes_to_list(x) for x in self.ff.angletypes.values()],
        )

    def update_parameters(self):
        if self.needs_parameterization:
            if self.parametrizer is not None:
                focus_ids = self.parameterization_focus_ids
                logger.info(
                    f"Starting parametrization using {self.parametrizer.__class__.__name__} with focus on {focus_ids}"
                )
                self.parametrizer.parameterize_topology(self, focus_nrs=focus_ids)
            else:
                raise RuntimeError("No Parametrizer was initialized in this topology!")
            self.needs_parameterization = False

    # def update_partial_charges(self, recipe_steps: list[RecipeStep]) -> None:
    #     """Update the topology atom partial charges.

    #     This function must be called after the recipe_steps are applied.
    #     Changes are based on the recipe_steps. Update rules follow a simple
    #     assignment scheme that works well with grappa. If fragments are created,
    #     their partial charges are kept integers. If previously broken bonds are
    #     formed again, the original partial charges are restored.
    #     """

    #     # build updated list of atoms by residue for topology so that this does
    #     # not need to be repeated
    #     # Warning: resnr not unique for each residue, 'residues' can map to
    #     # more than one real residue
    #     residues = {}
    #     for atom in self.atoms.values():
    #         if residues.get(atom.resnr) is None:
    #             residues[atom.resnr] = []
    #         residues[atom.resnr].append(atom)

    #     for step in recipe_steps:
    #         if isinstance(step, Break):
    #             # make partial charges on either side of the break integer
    #             if self.atoms[step.atom_id_1].resnr == self.atoms[step.atom_id_2].resnr:
    #                 atom1 = self.atoms[step.atom_id_1]
    #                 atom2 = self.atoms[step.atom_id_2]
    #                 fragment1, fragment2 = get_residue_fragments(
    #                     self, residues[atom1.resnr], atom1, atom2
    #                 )
    #                 if len(fragment2) in (0, 1):
    #                     # intra-residue HAT case (or similar)
    #                     if atom1.type.upper().startswith("H"):
    #                         atom2.charge = (
    #                             f"{float(atom2.charge) + float(atom1.charge):7.4f}"
    #                         )
    #                         atom1.charge = "0.0"
    #                     elif atom2.type.upper().startswith("H"):
    #                         atom1.charge = (
    #                             f"{float(atom1.charge) + float(atom2.charge):7.4f}"
    #                         )
    #                         atom2.charge = "0.0"
    #                     else:
    #                         logger.warning(
    #                             "Residue is not fragmented and no break atom is hydrogen. Doing nothing!"
    #                         )
    #                         continue
    #                 else:
    #                     # homolysis case (or similar)

    #                     charge_fragment1 = [
    #                         float(self.atoms[nr].charge) for nr in fragment1
    #                     ]
    #                     charge_fragment2 = [
    #                         float(self.atoms[nr].charge) for nr in fragment2
    #                     ]
    #                     diff1 = sum(charge_fragment1) - round(sum(charge_fragment1))
    #                     diff2 = sum(charge_fragment2) - round(sum(charge_fragment2))
    #                     atom1.charge = f"{float(atom1.charge) - diff1:7.4f}"
    #                     atom2.charge = f"{float(atom2.charge) - diff2:7.4f}"
    #             else:
    #                 # no change in partial charges necessary for break between residues
    #                 continue
    #         elif isinstance(step, Bind):
    #             if self.atoms[step.atom_id_1].resnr == self.atoms[step.atom_id_2].resnr:
    #                 atom1 = self.atoms[step.atom_id_1]
    #                 atom2 = self.atoms[step.atom_id_2]
    #                 # check whether bond exists in topology
    #                 residue = atom1.residue
    #                 bondtype = (atom1.atom, atom2.atom)
    #                 residuetype = self.ff.residuetypes.get(residue)
    #                 if residuetype is None:
    #                     m = f"Residue {residue} not found in residuetypes. Can't update partial charges!"
    #                     logger.warning(m)
    #                     continue
    #                 residue_bond_spec = residuetype.bonds.get(
    #                     bondtype,
    #                     residuetype.bonds.get((bondtype[-1], bondtype[-2])),
    #                 )
    #                 if residue_bond_spec:
    #                     atom1.charge = residuetype.atoms[atom1.atom].charge
    #                     atom2.charge = residuetype.atoms[atom2.atom].charge
    #                 else:
    #                     logger.warning(
    #                         f"New bond defined in {step} but can't be found in residuetypes definition. Not changing the partial charges!"
    #                     )
    #             else:
    #                 # no change in partial charges necessary
    #                 continue
    #         else:
    #             continue

    def to_dict(self) -> TopologyDict:
        self.update_parameters()
        self._update_dict()
        return self.top

    # def del_atom(
    #     self, atom_nr: Union[list[str], str], parameterize: bool = True
    # ) -> dict[str, str]:
    #     """Deletes atom

    #     Deletes atom and all attached bonds. Reindexes the top and updates the
    #     parameters if requested. Also moves charges to first bound_nrs atom.

    #     Parameters
    #     ----------
    #     atom_nr
    #         1-based atom number as string to delete
    #     parameterize
    #         If true and bonds are removed triggers reparameterization,
    #         by default True

    #     Returns
    #     -------
    #     update_map
    #         Dict, mapping of old atom number strings to new ones.
    #     """
    #     if not isinstance(atom_nr, list):
    #         atom_nr = [atom_nr]
    #     for _atom_nr in atom_nr:
    #         atom = self.atoms[_atom_nr]
    #         logger.debug(
    #             f"Deleting Atom nr {atom.nr}, type {atom.type}, res {atom.residue}"
    #         )
    #         if len(atom.bound_to_nrs) > 0:
    #             # move charge to first neighbor
    #             self.atoms[atom.bound_to_nrs[0]].charge = (
    #                 f"{float(self.atoms[atom.bound_to_nrs[0]].charge) + float(atom.charge):7.4f}"
    #             )

    #             # break all bonds and delete all pairs, diheadrals with these bonds
    #             for bound_nr in copy(atom.bound_to_nrs):
    #                 self.break_bond((bound_nr, _atom_nr))

    #             for an in tuple(self.angles.keys()):
    #                 if _atom_nr in an:
    #                     self.angles.pop(an)

    #             for pd in tuple(self.proper_dihedrals.keys()):
    #                 if _atom_nr in pd:
    #                     self.proper_dihedrals.pop(pd)

    #             for id in tuple(self.improper_dihedrals.keys()):
    #                 if _atom_nr in id:
    #                     self.improper_dihedrals.pop(id)

    #         self.radicals.pop(_atom_nr)
    #         self.atoms.pop(_atom_nr)

    #     update_map_all = self.reindex_atomnrs()
    #     update_map = update_map_all[self.merged_moleculetype]

    #     if parameterize:
    #         self.update_parameters()
    #         # Overwriting in case of no parameterization wanted
    #         self.needs_parameterization = False
    #         self.parameterization_focus_ids = set()

    #     return update_map

    def reindex_atomnrs(self) -> dict[str, dict[str, str]]:
        """Reindex atom numbers in topology.

        Starts at index 1.
        This also updates the numbers for bonds, angles, dihedrals and pairs.
        Returns a dict of all moleculetypes to their update maps (old -> new).
        """
        update_map_all = {}
        for id, moleculetype in self.moleculetypes.items():
            update_map_all[id] = moleculetype.reindex_atomnrs()
        self._link_atomics()  # necessary because reindex_atomnrs creates a new dict
        return update_map_all

    def _regenerate_topology_from_bound_to(self):
        """Regenerate the topology from the bound_to lists of the atoms."""
        for moleculetype in self.moleculetypes.values():
            moleculetype._regenerate_topology_from_bound_to(self.ff)

    def __str__(self) -> str:
        molecules = "\n".join([name + ": " + n for name, n in self.molecules])
        selected_moleculetype = self.moleculetypes.get(self.selected_moleculetype)
        text = (
            textwrap.dedent(f"Topology with the following molecules: \n{molecules}\n\n")
            + textwrap.dedent(f"With {selected_moleculetype}")
            + textwrap.dedent(f"{self.ff}")
        )
        return text

    def __repr__(self) -> str:
        self._update_dict()
        return f"Topology({self.top})"

    def _repr_pretty_(self, p, _):
        """A __repr__ for ipython.

        This whill be used if just the name of the object is entered in the ipython shell
        or a jupyter notebook.

        p is an instance of
        [IPython.lib.pretty.PrettyPrinter](https://ipython.org/ipython-doc/3/api/generated/IPython.lib.pretty.html)
        """
        p.text(str(self))

    # def get_neighbors(self, initial_atoms: set[str], order: int):
    #     explored_atoms = initial_atoms
    #     frontier = initial_atoms
    #     for _ in range(order):
    #         new_frontier = set()
    #         for atom in frontier:
    #             new_frontier.update(self.atoms[atom].bound_to_nrs)
    #         frontier = new_frontier - explored_atoms
    #         explored_atoms |= new_frontier
    #     return explored_atoms

    # def break_bond(
    #     self, atompair_addresses: tuple[TopologyAtomAddress, TopologyAtomAddress]
    # ):
    #     """Break bonds in topology homolytically.

    #     Removes bond, angles and dihedrals where atompair was involved.
    #     Modifies the topology dictionary in place.
    #     Atom pairs become radicals.

    #     Parameters
    #     ----------
    #     atompair_addresses
    #         Between which atoms to break the bond.
    #     """
    #     logger.debug(f"Breaking bond between {atompair_addresses}")
    #     merged_moleculetype = self.merged_molecule

    #     # tuple -> list -> sorted -> tuple still makes it a tuple of two strings
    #     # so pyright can chill.
    #     atompair_nrs: tuple[str, str] = tuple(sorted(atompair_addresses, key=int))  # type: ignore

    #     atompair = [
    #         merged_moleculetype.atoms[atompair_nrs[0]],
    #         merged_moleculetype.atoms[atompair_nrs[1]],
    #     ]

    #     if not self.validate_bond(*atompair):
    #         raise ValueError(
    #             "Trying to break non-existing bond!"
    #             f"\n\tatom 1, nr {atompair[0].nr}, type {atompair[0].type}, res {atompair[0].residue}"
    #             f"\n\tatom 2, nr {atompair[1].nr}, type {atompair[1].type}, res {atompair[1].residue}"
    #         )

    #     self.needs_parameterization = True
    #     self.parameterization_focus_ids.update(atompair_nrs)

    #     # mark atoms as radicals
    #     for atom in atompair:
    #         atom.is_radical = True
    #         merged_moleculetype.radicals[atom.nr] = atom

    #     # bonds
    #     # remove bond
    #     removed_bond = merged_moleculetype.bonds.pop(atompair_nrs)
    #     logging.debug(f"removed bond: {removed_bond}")

    #     # remove angles
    #     angle_keys = merged_moleculetype._get_center_atom_angles(
    #         atompair_nrs[0]
    #     ) + merged_moleculetype._get_center_atom_angles(atompair_nrs[1])
    #     for key in angle_keys:
    #         if all([x in key for x in atompair_nrs]):
    #             # angle contained a now deleted bond because
    #             # it had both atoms of the broken bond
    #             merged_moleculetype.angles.pop(key, None)

    #     # remove proper dihedrals
    #     # and pairs
    #     dihedral_keys = merged_moleculetype._get_atom_proper_dihedrals(
    #         atompair_nrs[0]
    #     ) + merged_moleculetype._get_atom_proper_dihedrals(atompair_nrs[1])
    #     for key in dihedral_keys:
    #         # don't use periodicity in key for checking atompair_nrs
    #         if all([x in key for x in atompair_nrs]):
    #             # dihedral contained a now deleted bond because
    #             # it had both atoms of the broken bond
    #             merged_moleculetype.proper_dihedrals.pop(key, None)
    #             pairkey: tuple[str, str] = tuple(sorted((key[0], key[3]), key=int))  # type: ignore
    #             merged_moleculetype.pairs.pop(pairkey, None)

    #     # and improper dihedrals
    #     keys_remove = []
    #     for key in merged_moleculetype.improper_dihedrals:
    #         if all([x in key for x in atompair_nrs]):
    #             keys_remove.append(key)
    #     for key in keys_remove:
    #         merged_moleculetype.improper_dihedrals.pop(key, None)

    #     # warn if atom is part of restraint
    #     for atom in atompair:
    #         posres = atom.nr in self.position_restraints.keys()
    #         dihres = any([atom.nr in atms for atms in self.position_restraints.keys()])
    #         if posres or dihres:
    #             logging.warning(
    #                 "Deleting bond with atom involved in a restraint!"
    #                 "This may causes unwanted behaviour.\n"
    #                 f"\n\tatom 1, nr {atompair[0].nr}, type {atompair[0].type}, res {atompair[0].residue}"
    #                 f"\n\tatom 2, nr {atompair[1].nr}, type {atompair[1].type}, res {atompair[1].residue}"
    #             )

    #     # update bound_to
    #     atompair[0].bound_to_nrs.remove(atompair[1].nr)
    #     atompair[1].bound_to_nrs.remove(atompair[0].nr)

    # def bind_bond(
    #     self, atompair_addresses: tuple[TopologyAtomAddress, TopologyAtomAddress]
    # ):
    #     """Add a bond in topology.

    #     Modifies the topology dictionary in place.
    #     It keeps track of affected terms in the topology via a graph representation of the topology
    #     and applies the necessary changes to bonds, angles and dihedrals (proper and improper).
    #     Furthermore, it modifies to function types in the topology to account for radicals.

    #     Parameters
    #     ----------
    #     atompair_addresses
    #         Atoms to bind together.
    #     """
    #     logger.debug(f"Creating bond between {atompair_addresses}")

    #     merged_moleculetype = self.merged_molecule

    #     atompair_nrs: tuple[str, str] = tuple(sorted(atompair_addresses, key=int))  # type: ignore
    #     atompair = [
    #         merged_moleculetype.atoms[atompair_nrs[0]],
    #         merged_moleculetype.atoms[atompair_nrs[1]],
    #     ]

    #     # check whether they are bound already
    #     if self.validate_bond(*atompair):
    #         raise ValueError(
    #             "Trying to bind to atoms already bound!"
    #             f"\n\tatom 1, nr {atompair[0].nr}, type {atompair[0].type}, res {atompair[0].residue}"
    #             f"\n\tatom 2, nr {atompair[1].nr}, type {atompair[1].type}, res {atompair[1].residue}"
    #         )

    #     self.needs_parameterization = True
    #     self.parameterization_focus_ids.update(atompair_nrs)

    #     # de-radialize if re-combining two radicals
    #     if all(map(lambda x: x.is_radical, atompair)):
    #         atompair[0].is_radical = False
    #         atompair[1].is_radical = False
    #         for a in atompair:
    #             if merged_moleculetype.radicals.pop(a.nr) is None:
    #                 logging.warning(
    #                     f"Atom {a} should have been a radical but was "
    #                     "not registed in moleculetype.radicals"
    #                 )

    #     # TODO: Do we still need this with grappa?
    #     # quickfix for jumping hydrogens
    #     # to make them adopt the correct residuetype and atomtype
    #     # when bound to a new heavy atom
    #     for i, atom in enumerate(atompair):
    #         if atom.type.upper().startswith("H"):
    #             other_i = abs(i - 1)
    #             other_atom = atompair[other_i]
    #             other_res = other_atom.residue
    #             atom.residue = other_res
    #             atom.resnr = other_atom.resnr
    #             aa = self.ff.residuetypes.get(other_res)
    #             if not aa:
    #                 logging.warning(
    #                     f"No residuetype found in ff for {other_res}, to which a H would bind"
    #                 )
    #                 continue

    #             other_residue = get_residue_by_bonding(
    #                 other_atom, merged_moleculetype.atoms
    #             )
    #             residue_atomnames_current = [a.atom for a in other_residue.values()]
    #             name_set = False
    #             for key, bond in aa.bonds.items():
    #                 if other_atom.atom in key and any(
    #                     k.upper().startswith("H") for k in key
    #                 ):
    #                     if key[0] == other_atom.atom:
    #                         h_name = bond.atom2
    #                     else:
    #                         h_name = bond.atom1

    #                     atom.type = aa.atoms[h_name].type
    #                     if (
    #                         not h_name in residue_atomnames_current
    #                         or h_name == atom.atom
    #                     ):
    #                         atom.atom = h_name
    #                         name_set = True
    #                     else:
    #                         atom.atom = "HX"
    #                         name_set = True
    #                         continue
    #                     logger.debug(f"Hydrogen will be bound to {other_atom}.")
    #                     break
    #             else:
    #                 if name_set:
    #                     logger.warning(
    #                         "Found new atomtype but not atomname for HAT hydrogen!"
    #                     )
    #                 else:
    #                     logger.warning(
    #                         "Found neither new atomtype nor atomname for HAT hydrogen!"
    #                     )
    #             if not name_set:
    #                 atom.atom = "HX"
    #                 logger.debug(f"Named newly bonded hydrogen 'HX'")

    #     # update bound_to
    #     atompair[0].bound_to_nrs.append(atompair[1].nr)
    #     atompair[1].bound_to_nrs.append(atompair[0].nr)

    #     # bonds
    #     # add bond
    #     bond = Bond(atompair_nrs[0], atompair_nrs[1], FFFUNC["harmonic_bond"])
    #     merged_moleculetype.bonds[atompair_nrs] = bond
    #     logging.info(f"added bond: {bond}")

    #     # make sure that there are no additional pairs
    #     # where bonds or angles are already between the atoms
    #     bond_pairkey: tuple[str, str] = tuple(
    #         str(x) for x in sorted([atompair[0].nr, atompair[1].nr], key=int)
    #     )  # type: ignore
    #     merged_moleculetype.pairs.pop(bond_pairkey, None)

    #     # and remove improper dihedrals connected to the bonds,
    #     # as they will be re-added based on potentially changed residuetypes
    #     to_pop = []
    #     for key in merged_moleculetype.improper_dihedrals.keys():
    #         if atompair_nrs[0] in key or atompair_nrs[1] in key:
    #             to_pop.append(key)

    #     for key in to_pop:
    #         merged_moleculetype.improper_dihedrals.pop(key, None)

    #     # add angles
    #     angle_keys = merged_moleculetype._get_atom_angles(
    #         atompair_nrs[0]
    #     ) + merged_moleculetype._get_atom_angles(atompair_nrs[1])
    #     for key in angle_keys:
    #         if merged_moleculetype.angles.get(key) is None:
    #             merged_moleculetype.angles[key] = Angle(
    #                 key[0], key[1], key[2], FFFUNC["harmonic_angle"]
    #             )
    #         # make sure that there are no additional pairs
    #         # where bonds or angles are already between the atoms
    #         angle_pairkey: tuple[str, str] = tuple(
    #             str(x) for x in sorted([key[0], key[2]], key=int)
    #         )  # type: ignore
    #         merged_moleculetype.pairs.pop(angle_pairkey, None)

    #     # add proper and improper dihedrals
    #     # add proper dihedrals and pairs
    #     # this assumes function type 9 for all proper dihedrals
    #     # and adds all possible dihedrals (and pairs)
    #     dihedral_keys = merged_moleculetype._get_atom_proper_dihedrals(
    #         atompair_nrs[0]
    #     ) + merged_moleculetype._get_atom_proper_dihedrals(atompair_nrs[1])
    #     for key in dihedral_keys:
    #         if merged_moleculetype.proper_dihedrals.get(key) is None:
    #             merged_moleculetype.proper_dihedrals[key] = MultipleDihedrals(
    #                 *key,
    #                 FFFUNC["mult_proper_dihedral"],
    #                 dihedrals={"": Dihedral(*key, FFFUNC["mult_proper_dihedral"])},
    #             )
    #         pairkey: tuple[str, str] = tuple(
    #             str(x) for x in sorted([key[0], key[3]], key=int)
    #         )  # type: ignore
    #         if merged_moleculetype.pairs.get(pairkey) is None:
    #             # only add pair if the atoms are not otherwise bound (in a circle)
    #             if (
    #                 pairkey[0]
    #                 not in merged_moleculetype.atoms[pairkey[1]].bound_to_nrs
    #             ):
    #                 merged_moleculetype.pairs[pairkey] = Pair(
    #                     pairkey[0], pairkey[1], FFFUNC["pair"]
    #                 )

    #     # improper dihedral
    #     atom_impropers = merged_moleculetype._get_atom_improper_dihedrals(
    #         atompair_nrs[0], self.ff
    #     )
    #     atom_impropers.update(
    #         merged_moleculetype._get_atom_improper_dihedrals(atompair_nrs[1], self.ff)
    #     )
    #     for key, value in atom_impropers.items():
    #         if value.c2 is None:
    #             value.c2 = ""
    #         if merged_moleculetype.improper_dihedrals.get(key) is None:
    #             merged_moleculetype.improper_dihedrals[key] = MultipleDihedrals(
    #                 *key, FFFUNC["mult_improper_dihedral"], dihedrals={}
    #             )
    #         merged_moleculetype.improper_dihedrals[key].dihedrals[value.c2] = (
    #             Dihedral(
    #                 *key,
    #                 FFFUNC["mult_improper_dihedral"],
    #                 c0=value.c0,
    #                 c1=value.c1,
    #                 periodicity=value.c2,
    #             )
    #         )

    #     # remove settles and explicit exclusions
    #     # those are used by solvent molecules
    #     # but if a solvent molecule get's bound to other parts
    #     # of the merged molecule it is no longer a solvent molecule
    #     for ai in atompair_nrs:
    #         to_delete = []
    #         for exclusion_key in merged_moleculetype.exclusions.keys():
    #             if ai in exclusion_key:
    #                 to_delete.append(exclusion_key)

    #         if len(to_delete) > 0:
    #             logger.debug(f"Removing exclusions {to_delete}")
    #         for key in to_delete:
    #             merged_moleculetype.exclusions.pop(key)
    #         settles = merged_moleculetype.settles.get(ai)
    #         if settles is not None:
    #             logger.debug(f"Removing settles {ai}")
    #             merged_moleculetype.settles.pop(ai)



# This file is part of the gmx-top4py project.
#
# The gmx-top4py project is based on or includes code from:
#    kimmdy (https://github.com/graeter-group/kimmdy/tree/main)
#    Copyright (C) graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# Modifications and additional code:
#    Copyright (C) 2025 graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Utilities for shell convenience functions and GROMACS related functions
"""

import logging
import subprocess as sp
from pathlib import Path
from typing import Optional


logger = logging.getLogger(__name__)

TopologyAtomAddress = str
"""Address to an atom in the topology.

Corresponds to the 1-based id in the topology and coordinates file.
Note, that gromacs ids in the atomnr column of the gro file
can overflow due to the fixed width file format.
The line number - 2 (for the title and the number of atoms) is always
the correct the atom id.
"""


def field_or_none(l: list[str], i) -> Optional[str]:
    try:
        return l[i]
    except IndexError as _:
        return None


### GROMACS related functions ###


def get_gmx_dir(
    gromacs_alias: str = "gmx", grompp_prefix: Optional[str] = None
) -> Optional[Path]:
    """Returns the path to the gromacs installation"""

    # get the stder from calling `gmx` to search for the `Data prefix:`
    # line which contains the path to the gromacs installation

    # Add prefix if necesarry
    cmd = [gromacs_alias]
    if grompp_prefix:
        cmd.insert(0, grompp_prefix)
    try:
        r = sp.run(cmd, check=False, capture_output=True, text=True)
    except FileNotFoundError:
        logger.warning("GROMACS not found.")
        return None

    from_source = False
    gmx_prefix = None
    for l in r.stderr.splitlines():
        if l.startswith("Data prefix:"):
            gmx_prefix = Path(l.split()[2])
            if "(source tree)" in l:
                from_source = True
            break

    if gmx_prefix is None:
        logger.warning("GROMACS data directory not found in gromacs message.")
        return None

    if from_source:
        return Path(gmx_prefix) / "share"
    else:
        return Path(gmx_prefix) / "share" / "gromacs"



"""
Welcome to gmx-top4py!
"""

from __future__ import annotations
import warnings

warnings.simplefilter("once", DeprecationWarning)
warnings.warn(
    "gmx-top4py has been renamed to gmxtop. Use 'pip install gmxtop'.",
    DeprecationWarning,
)



# This file is part of the gmx-top4py project.
#
# The gmx-top4py project is based on or includes code from:
#    kimmdy (https://github.com/graeter-group/kimmdy/tree/main)
#    Copyright (C) graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# Modifications and additional code:
#    Copyright (C) 2025 graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
All read_<...> and write_<...> functions.
"""

import csv
import json
import logging
import os
from datetime import datetime
from itertools import takewhile
from pathlib import Path
from typing import Optional, TypeAlias, TypedDict, Union
import numpy as np

from gmx_top4py.utils import get_gmx_dir

logger = logging.getLogger(__name__)
TopologyDict = dict
"""A raw representation of a topology file returned by [](`~gmx-top4py.parsing.read_top`).
"""


## convenience functions
def is_not_comment(c: str) -> bool:
    """Returns whether a string is not a comment.

    Used for topology like files that use ';' for comments.
    """
    return c != ";"


def resolve_includes(
    path: Path, gmx_builtin_ffs: Optional[Path] = None
) -> tuple[list[str], Optional[Path]]:
    """Resolve #include statements in a (top/itp) file.

    Parameters
    ----------
    path :
        Filepath to read.

    Returns
    -------
    lines:
        List of lines.
    ffdir:
        Path to the ff directory if one of the includes used a file from it.
    """
    path = path.absolute()
    dir = path.parent
    cwd = Path.cwd()
    if not dir.exists() or not path.exists():
        return ([], None)
    os.chdir(dir)
    ls = []
    ffdir = None

    with open(path, "r") as f:
        for l in f:
            l = "".join(takewhile(is_not_comment, l)).strip()
            if not l:
                continue
            if l.startswith("#include"):
                include_path = Path(l.split('"')[1])
                # if the include path contains `ff` it is a file in a force field directory
                # get the path to the force field directory
                if ".ff" in l:
                    # e.g. #include "amber99.ff/forcefield.itp"
                    ffdir = include_path.parent
                    # test if the path is in the cwd, otherwise search in gmx ff dir
                    if not ffdir.exists() and gmx_builtin_ffs is not None:
                        ffdir = gmx_builtin_ffs / ffdir
                    ffdir = ffdir.absolute()
                ls_prime, _ = resolve_includes(include_path, gmx_builtin_ffs)
                if not ls_prime and gmx_builtin_ffs is not None:
                    ls_prime, _ = resolve_includes(
                        gmx_builtin_ffs / include_path, gmx_builtin_ffs
                    )
                if not ls_prime:
                    logger.info(
                        f"top include {include_path} could not be resolved. Line was dropped."
                    )
                ls.extend(ls_prime)
            else:
                ls.append(l)

    os.chdir(cwd)
    return (ls, ffdir)


## parsing functions ##
## topology
def empty_section(condition: Optional[dict] = None) -> dict:
    return {"content": [], "else_content": [], "extra": [], "condition": condition}


def read_top(
    path: Path,
    ffdir: Optional[Path] = None,
    use_gmx_dir: bool = True,
) -> TopologyDict:
    """Read a topology file (*.top,*.itp,*.rtp) into a raw TopologyDict represenation.

    Assumptions and limitation:

    - `#include` statements will be resolved
    - comments will be removed
    - all lines are stripped of leading and trailing whitespace
    - `#undef` is not supported
    - a section within `ifdef` may be a subsection of a section that was started
      outside of the `ifdef`
    - `#if..#endif` statements only surround a full section or subsection,
      not individual lines within a section and
      a section may either be contained within if ... else or it may not be,
      but it can not be duplicated with one part inside and one outside.
    - `if .. else` can't be nested
    - ``#include`` s that don't resolve to a valid file path are silently dropped
    - sections that can have subsections can also exist multiple, separate times
      e.g. moleculetype will appear multiple times and they should not be merged

    Parameters
    ----------
    path :
        Path to the topology file.

    Returns
    ------
    Every section, apart from `ffdir` and `define`,
    comes with a condition that can be checked against the
    `define`s by the helper functions to determine if the content
    (a list of lists) should come from `content` or `else_content`.
    Some sections such as `moleculetype` also come with `subsections`.

    Examples
    -------
    ```python
    raw_top =
    {'ffdir': PosixPath('/usr/share/gromacs/top/amber99.ff'),
    'define': {'_FF_AMBER': [], '_FF_AMBER99': []},
    'defaults': {'content': [['1', '2', 'yes', '0.5', '0.8333']],
    'else_content': [],
    'extra': [],
    'condition': None},
    'atomtypes': {'content': [
    ['C', '6', '12.01', '0.0000', 'A', '3.39967e-01', '3.59824e-01'],
    ['MNH3', '0', '0.0000', '0.0000', 'A', '0.00000e+00', '0.00000e+00']],
    'else_content': [],
    'extra': [],
    'condition': None},
    'moleculetype_Urea': {'content': [['Urea', '3']],
    'else_content': [],
    'extra': [],
    'condition': None,
    'subsections': {'atoms': {'content': [['1',
        'C',
        '1',
        'URE',
        'C',
        '1',
        '0.880229',
        '12.01000'],
        ['2', 'O', '1', 'URE', 'O', '2', '-0.613359', '16.00000'],
        ['3', 'N', '1', 'URE', 'N1', '3', '-0.923545', '14.01000'],
        ['4', 'H', '1', 'URE', 'H11', '4', '0.395055', '1.00800'],
        ['5', 'H', '1', 'URE', 'H12', '5', '0.395055', '1.00800'],
        ['6', 'N', '1', 'URE', 'N2', '6', '-0.923545', '14.01000'],
        ['7', 'H', '1', 'URE', 'H21', '7', '0.395055', '1.00800'],
        ['8', 'H', '1', 'URE', 'H22', '8', '0.395055', '1.00800']],
        'else_content': [],
        'extra': [],
        'condition': None},
    'bonds': {'content': [['1', '2'],
        ['1', '3'],
        ['1', '6'],
        ['3', '4'],
        ['3', '5'],
        ['6', '7'],
        ['6', '8']],
        'else_content': [],
        'extra': [],
        'condition': None},
    'dihedrals': {'content': [['2', '1', '3', '4', '9'],
        ['2', '1', '3', '5', '9'],
        ['2', '1', '6', '7', '9'],
        ['2', '1', '6', '8', '9'],
        ['3', '1', '6', '7', '9'],
        ['3', '1', '6', '8', '9'],
        ['6', '1', '3', '4', '9'],
        ['6', '1', '3', '5', '9'],
        ['3', '6', '1', '2', '4'],
        ['1', '4', '3', '5', '4'],
        ['1', '7', '6', '8', '4']],
        'else_content': [],
        'extra': [],
        'condition': None},
    'position_restraints': {'content': [['1', '1', '1000', '1000', '1000'],
        ['2', '1', '1000', '0', '1000'],
        ['3', '1', '1000', '0', '0']],
        'else_content': [],
        'extra': [],
        'condition': None},
    'dihedral_restraints': {'content': [['3',
        '6',
        '1',
        '2',
        '1',
        '180',
        '0',
        '10'],
        ['1', '4', '3', '5', '1', '180', '0', '10']],
        'else_content': [],
        'extra': [],
        'condition': None}}},
    }
    ```
    """

    CHILD_SECTIONS = (
        "atoms",
        "bonds",
        "pairs",
        "angles",
        "dihedrals",
        "impropers",
        "pairs_nb",
        "exclusions",
        "constraints",
        "virtual_sites",
        "virtual_sites1",
        "virtual_sites2",
        "virtual_sites3",
        "virtual_sites4",
        "virtual_sitesn",
        "settles",
        "position_restraints",
        "dihedral_restraints",
        "angle_restraints",
        "angle_restraints_z",
        "orientation_restraints",
    )
    if not path.is_file():
        raise FileNotFoundError(f"topology file {path} not found")
    gmx_builtin_ffs = None
    if use_gmx_dir:
        gmxdir = get_gmx_dir()
        if gmxdir is not None:
            gmx_builtin_ffs = gmxdir / "top"
    ls, parsed_ffdir = resolve_includes(path, gmx_builtin_ffs)
    if ffdir is None and parsed_ffdir is not None:
        ffdir = parsed_ffdir

    if ffdir is None:
        # try to find a forcefield directory in the parent directory of the topology
        # or in the current working directory
        # with the *.ff glob
        dir = path.parent
        cwd = Path.cwd()
        dirs = list(dir.glob("*.ff"))
        if dirs:
            ffdir = dirs[0]
            logger.debug(f"Found forcefield directory {ffdir} in {path.parent}")
        else:
            # check the current working directory
            dirs = list(cwd.glob("*.ff"))
            if dirs:
                ffdir = dirs[0]
                logger.debug(f"Found forcefield directory {ffdir} in {cwd}")
            else:
                ffdir = None

    if ffdir is None:
        logger.debug(
            f"No #include for a forcefield directory found in {path} and no directory *.ff found in parent directory of the topology or the cwd."
        )

    ls = [l for l in ls if not l.startswith("*")]
    d = {}
    d["ffdir"] = ffdir
    d["define"] = {}
    parent_section = None
    section = None
    condition = None
    is_first_line_after_section_header = False
    content_key = "content"

    for i, l in enumerate(ls):
        # decide where to put lines depending on current context
        # deal with statements
        if l.startswith("#define"):
            l = l.split()
            name = l[1]
            values = l[2:]
            d["define"][name] = values
            continue
        elif l.startswith("#if"):
            if is_first_line_after_section_header:
                raise NotImplementedError(
                    f"""Error parsing {path}: #if ... #endif can only be used to surround a section, not within."""
                )
            l = l.split()
            condition_type = l[0].removeprefix("#")
            condition_value = l[1]
            condition = {"type": condition_type, "value": condition_value}
            continue
        elif l.startswith("#else"):
            content_key = "else_content"
            continue
        elif l.startswith("#endif"):
            condition = None
            content_key = "content"
            continue
        # deal with a new section
        elif l.startswith("["):
            section = l.strip("[] \n")
            # add nested structure for non-nestable sections
            if section not in CHILD_SECTIONS:
                parent_section = section
                section = None
                if d.get(parent_section) is None:
                    if parent_section == "moleculetype":
                        # append the name of the moleculetype to the section name
                        # but the name comes on the next line
                        moleculetype_name = ls[i + 1].split()[0]
                        parent_section = f"{parent_section}_{moleculetype_name}"
                    d[parent_section] = empty_section(condition)
                    d[parent_section]["subsections"] = {}
            else:
                if parent_section is not None:
                    # add empty section as subsection
                    if d[parent_section]["subsections"].get(section) is None:
                        d[parent_section]["subsections"][section] = empty_section(
                            condition
                        )
                else:
                    # add regular section
                    if d.get(section) is None:
                        d[section] = empty_section(condition)
            is_first_line_after_section_header = True
        # deal with content lines
        else:
            if parent_section is not None:
                # line is in a section that can have subsections
                if section is None:
                    # but no yet in a subsection
                    l = l.split()
                    d[parent_section][content_key].append(l)
                else:
                    # part of a subsection
                    d[parent_section]["subsections"][section][content_key].append(
                        l.split()
                    )
            else:
                if section is None:
                    raise ValueError(
                        f"topology file {path} contains lines outside of a section"
                    )
                # adding content to dict here; this is where the loop goes 98% of the time
                d[section][content_key].append(l.split())
            is_first_line_after_section_header = False

    if len(d) <= 2:
        m = f"topology file {path} does not contain any sections"
        logger.error(m)
        raise ValueError(m)

    return d


def write_top(top: TopologyDict, outfile: Path):
    """Write a TopologyDict to a topology file.

    Parameters
    ----------
    top:
        Raw dictionary represenation of the topology.
    outfile :
        Path to the topology file to write to.
    """

    def write_if_nonempty(f, printname, ls):
        if ls:
            f.write(f"[ {printname} ]\n")
            for l in ls:
                f.writelines(" ".join(l))
                f.write("\n")

    def write_section(f, name: str, section):
        if name.startswith("moleculetype_"):
            printname = "moleculetype"
        else:
            printname = name
        condition = section.get("condition")
        f.write("\n")
        if condition is None:
            write_if_nonempty(f, printname, section["content"])
        else:
            condition_type = condition["type"]
            condition_value = condition["value"]
            f.write(f"#{condition_type} {condition_value}")
            f.write("\n")
            write_if_nonempty(f, printname, section["content"])
            else_content = section.get("else_content")
            if else_content:
                f.write("#else\n")
                write_if_nonempty(f, printname, else_content)
            f.write("#endif\n")

    with open(outfile, "w") as f:
        # write defines first
        define = top.get("define")
        if define is not None:
            for name, value in define.items():
                f.writelines("#define " + name + " ".join(value))
                f.write("\n")

        for name, section in top.items():
            if name in ["define", "ffdir", "system", "molecules"]:
                continue
            f.write("\n")
            subsections = section.get("subsections")
            write_section(f, name, section)
            if subsections is not None:
                for name, section in subsections.items():
                    write_section(f, name, section)

        # write system and molecules last
        for name in ["system", "molecules"]:
            section = top.get(name)
            if section is not None:
                f.write("\n")
                write_section(f, name, section)



# This file is part of the gmx-top4py project.
#
# The gmx-top4py project is based on or includes code from:
#    kimmdy (https://github.com/graeter-group/kimmdy/tree/main)
#    Copyright (C) graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# Modifications and additional code:
#    Copyright (C) 2025 graeter-group
#    Licensed under the GNU General Public License v3.0 (GPLv3).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
Constants used throughout gmx-top4py.
"""

ATOM_ID_FIELDS = {
    "atoms": [0, 5],  # atomnr, chargegroup
    "bonds": [0, 1],
    "angles": [0, 1, 2],
    "dihedrals": [0, 1, 2, 3],
    "pairs": [0, 1],
    "settles": [0],
    "exclusions": True,  # all fields at atom ids
    "position_restraints": [0],
    "dihedral_restraints": [0, 1, 2, 3],
}

# see https://manual.gromacs.org/current/reference-manual/topologies/topology-file-formats.html
FFFUNC = {
    "mult_proper_dihedral": "9",
    "mult_improper_dihedral": "4",
    "harmonic_bond": "1",
    "morse_bond": "3",
    "harmonic_angle": "1",
    "pair": "1",
    "coulomb_pair": "2",
}

RESNR_ID_FIELDS = {
    "atoms": [2],
}

# REACTIVE_MOLECULEYPE = "Reactive"


SOLVENT_NAMES: list[str] = [
    "SOL",
    "WATER",
    "TIP3P",
    "TIP4P",
    "TIP4P-Ew",
    "SPC",
    "SPC/E",
]

ION_NAMES: list[str] = [
    "I",
    "F",
    "CA",
    "CL",
    "NA",
    "MG",
    "K",
    "RB",
    "CS",
    "LI",
    "ZN",
]

# compare to atom type perception paper (2006) doi:10.1016/j.jmgm.2005.12.005
ATOMTYPE_BONDORDER: dict[tuple, int]
"""
To determin if an atom is a radical.
Compare to atom type perception paper (2006) doi:10.1016/j.jmgm.2005.12.005
"""
ATOMTYPE_BONDORDER = {
    ("MG", "NA", "CO"): 0,
    (
        "H",
        "HW",
        "HO",
        "HS",
        "HA",
        "HC",
        "H1",
        "H2",
        "H3",
        "HP",
        "H4",
        "H5",
        "HO",
        "H0",
        "HP",
        "O",
        "O2",
        "Cl",
        "Na",
        "I",
        "F",
        "Br",
    ): 1,
    ("NB", "NC", "OW", "OH", "OS", "SH", "S"): 2,
    (
        "C",
        "CN",
        "CB",
        "CR",
        "CK",
        "CC",
        "CW",
        "CV",
        "C*",
        "CQ",
        "CM",
        "CA",
        "CD",
        "CZ",
        "N",
        "NA",
        "N*",
        "N2",
    ): 3,
    ("CT", "N3", "P", "SO"): 4,
}


ATOMTYPE_BONDORDER_FLAT: dict[str, int]
"""
To determin if an atom is a radical.
Compare to atom type perception paper (2006) doi:10.1016/j.jmgm.2005.12.005
"""
ATOMTYPE_BONDORDER_FLAT = {
    "MG": 0,
    "NA": 0,
    "CO": 0,
    "H": 0,
    "HW": 0,
    "HO": 0,
    "HS": 0,
    "HA": 0,
    "HC": 0,
    "H1": 0,
    "H2": 0,
    "H3": 0,
    "HP": 0,
    "H4": 0,
    "H5": 0,
    "HO": 0,
    "H0": 0,
    "HP": 0,
    "O": 0,
    "O2": 0,
    "Cl": 0,
    "Na": 0,
    "I": 0,
    "F": 0,
    "Br": 0,
    "NB": 2,
    "NC": 2,
    "OW": 2,
    "OH": 2,
    "OS": 2,
    "SH": 2,
    "S": 2,
    "C": 3,
    "CN": 3,
    "CB": 3,
    "CR": 3,
    "CK": 3,
    "CC": 3,
    "CW": 3,
    "CV": 3,
    "C*": 3,
    "CQ": 3,
    "CM": 3,
    "CA": 3,
    "CD": 3,
    "CZ": 3,
    "N": 3,
    "NA": 3,
    "N*": 3,
    "N2": 3,
    "CT": 4,
    "N3": 4,
    "P": 4,
    "SO": 4,
}

AA3 = [
    "ALA",
    "CYS",
    "ASP",
    "GLU",
    "PHE",
    "GLY",
    "HIE",
    "ILE",
    "LYS",
    "LEU",
    "MET",
    "ASN",
    "PRO",
    "HYP",
    "GLN",
    "ARG",
    "SER",
    "THR",
    "VAL",
    "TRP",
    "TYR",
    "DOP",
    "ACE",
    "NME",
]
